/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#define AGENT_C
#include "bf_agent.h"

int     agent_re_exec = 0;
int     win_sock = 0;
char *  mapStr = NULL;
char *  defMapStr = NULL;
int     agent_mode = SE_OK;
int     ifd = 0;
int     ofd = 1;
char *  trapCommand = NULL;
char *  clearcase_view = NULL;
int     verbose = 1;
FILE *  LogF = NULL;
int     env_recursion_limit = 32;
char *  bf_debug = NULL;
int     heartbeat_interval = 15;
int     disable_telnet_support = 0;
char *  sourcelist = NULL;

bf_command_state_t command_state;

extern  int     did_hello;
extern  int     datablock_cache_size;
extern  char *  datablock_cache;

extern  const char *AGENT_VERSION;
extern  const char *SUPPORTED_ENV_ACTIONS;

bf_command_t commands[] = {
	{ "exec",          &FuncExec,      CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ "expand",        &FuncExpand,    CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ "read",          &FuncRead,      CMD_FLAG_CC | CMD_FLAG_ENV },
	{ "write",         &FuncWrite,     CMD_FLAG_CC | CMD_FLAG_ENV },
	{ "mkdir",         &FuncMkdir,     CMD_FLAG_CC | CMD_FLAG_ENV },
	{ "rmdir",         &FuncRmdir,     CMD_FLAG_CC | CMD_FLAG_ENV },
	{ "unlink",        &FuncUnlink,    CMD_FLAG_CC | CMD_FLAG_ENV },
	{ "ping",          &FuncPing,      CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ "sysconf",       &FuncSysConf,   CMD_FLAG_CC | CMD_FLAG_NO_REEXEC },
	{ "cc_changes",    &FuncCCChanges, CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ "showmethegopi", &FuncDumpConf,  CMD_FLAG_NO_REEXEC },
	{ NULL, NULL, 0 }
};


/* 
 *  function to output a message of some sort. 
 *  used while reading the request from stdin. 
 */
static void agent_send_message(int code, const char *message) {
	char buf[BFBUFSIZ];
	int len = sprintf(buf, "%d ", code);
	len += format_message(&buf[len], message, AGENT_VERSION);
	buf[len++] = '\n';
	buf[len] = '\0';
	io_puts(buf);
}

/* Become the requested user, if defined.  We also store the credentials,
 * for use later when we map drives.
 */
char *auth_user = NULL;
char *auth_pass = NULL;

static int agent_authenticate(void)
{
	int 	rc;
	char	buf[BFBUFSIZ], prob[BFBUFSIZ];
	const char *username = request_get(TAG_USERNAME);
	const char *password = request_get(TAG_ENC_PASS);
#ifdef WIN32
	char    *user = NULL;
#endif

	if (bf_debug) {
		sprintf(bf_debug, "agent_authenticate: username=%s password=%s",
				(username != NULL) ? username : "(null)",
				(password != NULL) ? "******" : "(null)");
		DebugBlock(bf_debug);
	}

#ifdef WIN32
	/* On windows, we first use a temporary buffer to convert / to \. */
	if (username != NULL) {
		char *p;
		user = strdup(username);
		if (NULL != (p = strchr(user, '/')))
			*p = '\\';
		username = user;
	}
#endif
	if (password == NULL)
		password = request_get(TAG_PASSWORD);

	/* If we authenticated before, allow anything except an attempt to
	 * use a different username.
	 */
	if (auth_user) {
		if (username == NULL || !*username || strcmp(username, auth_user) == 0)
			rc = SU_OK;
		else
			rc = SU_ILLEGAL;
	}
	else {
		rc = platform_set_user(username, password);
	}

	switch (rc) {
		case SU_OK:
			if (auth_user == NULL && username != NULL) {
				auth_user = strdup(username);
				if (password != NULL)
					auth_pass = strdup(password);
			}
			format_message(buf, bfCat(M_SETUSER_OK), auth_user);
			DataBlock(buf, NULL);
			rc = 0;
			break;

		case SU_OK_NOACCOUNT:
			rc = 0;
			break;

		case SU_ILLEGAL:
			format_message(prob, bfCat(E_SU_ILLEGAL),
					username ? username : bfCat(E_AUTH_NO_USER));
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_UNKNOWNUSER:
			format_message(prob, bfCat(E_SU_UNKNOWNUSER), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_ACCOUNTINVALID:
			format_message(prob, bfCat(E_SU_ACCOUNTINVALID), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_SETUSERFAILED:
			format_message(prob, bfCat(E_SU_SETUSERFAILED), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_USERNOTFOUND:
			format_message(prob, bfCat(E_SU_USERNOTFOUND), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_PAMSESSIONERROR:
			format_message(prob, bfCat(E_SU_PAMSESSIONERROR), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_NOCRED:
			format_message(prob, bfCat(E_SU_NOCRED), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		case SU_NOSESSION:
			format_message(prob, bfCat(E_SU_NOSESSION), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;

		default:
			format_message(prob, bfCat(E_SU_SYSTEMERROR), username);
			format_message(buf, bfCat(E_SETUSER_FAIL), prob, rc);
			ResultBlock(E_SETUSER_FAIL, -rc, buf);
			break;
	}

#ifdef WIN32
	/* On Windows, free the temporary buffer we used for
	 * converting / to \.
	 */
	if (user)
		free(user);
#endif
	return rc;
}

void ClobberPasswords(void) {
	if (bf_debug)
		DebugBlock("ClobberPasswords");
	if (auth_pass) {
		memset(auth_pass, 0, strlen(auth_pass));
		free(auth_pass);
		auth_pass = 0;
	}
	if (mapStr) memset(mapStr, 0, strlen(mapStr));
	request_del_creds();
}

static void agent_process_command(bf_command_t *cmd)
{
	bf_command_state_t *state = &command_state;
	const char *bfconf_file = bfconf_get_path();
	char *env = request_get(TAG_ENV);

	int i;
#ifndef WIN32
	int pty;
#endif

	memset(state, 0, sizeof(*state));
	state->mfd = state->sfd = -1;

	/* Turn on debugging if it was requested. */
	if (!bf_debug && request_has(TAG_DEBUG))
		bf_debug = calloc(1, BFBUFSIZ);

	if (bf_debug) {
		format_message(bf_debug, "agent_process_command: cmd=%s", cmd->cmd);
		DebugBlock(bf_debug);
	}

#define DEBUG_WIN_REEXEC 0

#ifdef WIN32
	/* Authenticate, unless in a re-exec (Windows only) */
	if (!agent_re_exec) {
		if (0 != (i = agent_authenticate()))
			return;
		if (!(cmd->flags & CMD_FLAG_NO_REEXEC) &&
				BFCONF_EXISTS(win_reexec_after_auth) &&
				(!IsOrigUser() || DEBUG_WIN_REEXEC))
		{
			platform_reexec_server();
			return;
		}
	}
#else /* !WIN32 */
	/* Always authenticate normally (non-Windows only) */
	if (0 != (i = agent_authenticate()))
		return;
#endif /* WIN32 */

	/* build the environment */
	if ((cmd->flags & CMD_FLAG_ENV) && NULL != env) {
		if (NULL != sourcelist) {
			free(sourcelist);
			sourcelist = NULL;
		}
		BuildEnv(env, &sourcelist); 
	}

#ifndef WIN32
	/* Allocate pseudo-tty / pipe (non-Windows only).  We need a pty
	 * if the command always needs one (exec/expand) or if we are using
	 * CC views, as the commands to mount them will need one.
	 */
	pty = cmd->flags & CMD_FLAG_PTY;
	if (!pty && (cmd->flags & CMD_FLAG_CC)) {
		char *p;
		if (NULL != (p = platform_getenv("_CLEARCASE_VIEWS")) ||
			NULL != (p = platform_getenv("_CLEARCASE_VOBS")) )
		{
			platform_freeenv(p);
			pty = 1;
		}
	}
	if (pty) {
		if (0 != (i = platform_openpty()))
			return;
	}
#endif /* WIN32 */


	/* Report any problems with the config file */
	if (0 != (i = bfconf_get_errno())) {
		format_message(state->buf, bfCat(E_CONFOPENFAIL), bfconf_file, i);
		DataBlock(state->buf, NULL);
	}


	/*
	 * Dispatch to the command.
	 */
	if (bf_debug)
		request_dump(bf_debug);
	(*cmd->func)();
	ClobberPasswords();



#ifndef WIN32
	/* Release pseudo-tty / pipe (non-Windows only) */
	if (pty) {
		if (state->mfd != -1) {
			close(state->mfd);
			state->mfd = -1;
		}
		if (state->sfd != -1) {
			close(state->sfd);
			state->sfd = -1;
		}
	}
#endif /* WIN32 */
}


/* 
 * Process a request block. this is basically a hash
 * of key=value pairs.
 */
int agent_process_request(void)
{
	const char * cmd;
	int          curr_cmd;

	if (bf_debug)
		DebugBlock("agent_process_request");

	if (NULL == (cmd = request_get(TAG_COMMAND))) {
		agent_send_message(CODE_NOCOMMAND, bfCat(CODE_NOCOMMAND));
		return 0;
	}

	/* locate the implementation of this command */
	for (curr_cmd = 0; commands[curr_cmd].cmd; ++curr_cmd) {
		if (0 == strcmp(commands[curr_cmd].cmd, cmd)) {
			/* Found a match; dispatch it! */
			agent_process_command(&commands[curr_cmd]);
			return 1;
		}
	}

	agent_send_message(CODE_BADCOMMAND, bfCat(CODE_BADCOMMAND));
	return 0;
}


static const char *TOKENS_EOL = "\r\n";
static const char *TOKENS_SPACE = " \t\r\n";

static void bfagent_main_loop(void) {
	char *bf_rptr = NULL, *bf_sptr = NULL;
	char *buf = malloc(BFBUFSIZ), *tag, *params;
	const char *p;

	/*  read stdin, firing off requests as we get them */
	while (io_gets(buf, BFBUFSIZ) != NULL) {
		/*  tokenize out the first word in the buffer. */
		/*  this will be the tag word or command.      */
		bf_strtok(NULL, NULL, NULL, &bf_sptr);
		tag = bf_strtok(buf, TOKENS_SPACE, &bf_rptr, &bf_sptr);
		tag = LowerString(tag);
		if (!tag || !*tag) { continue; }

		/* Discard lines with excessively long tags. */
		if (strlen(tag) > MAX_TAG_LEN) {
			io_ayt();
			break;
		}

		/* Time to quit? */
		if (strcmp(tag, TAG_QUIT) == 0) { break; }

		/* Set the language */
		if (strcmp(tag, TAG_LANG) == 0) {
			char *lang = bf_strtok(buf, TOKENS_SPACE, &bf_rptr, &bf_sptr);
			bfSetLang(lang);
			request_del(tag);
			request_put(tag, lang);
			continue;
		}

		/* Request is complete, so process it. */
		if (strcmp(tag, TAG_TRIGGER) == 0) {
			if (!agent_process_request()) { break; }
			CloseLog(); /* potentially open log file */
			env_clear();
			request_new();
			continue;
		}

		/* Make sure we don't need more entries than are allowed. */
		if (request_size() > MAX_REQUEST_TAGS && !request_has(tag)) {
			agent_send_message(E_TOOMANYTAGS, bfCat(E_TOOMANYTAGS));
			break;
		}

		/*  at this point we'll need any trailing args in the  */
		/*  input buffer. if there are none, set an empty string */
		/*  rather than a null pointer to make life easier later on. */
		params = bf_strtok(NULL, TOKENS_EOL, &bf_rptr, &bf_sptr);
		if (!params)
			params = "";

		/* the userdata tag is hex encoded so we can transport binary data */
		if (strcmp(tag, TAG_USERDATA) == 0) {
			size_t len = strlen(params) >> 1;
			unsigned char *value = calloc(1, len);
			DecodeBlock(value, (unsigned char *)params);
			request_put_binary(tag, value, len);
			continue;
		}

		/* The stream tag is used to record file data to a temporary file.
		 * If the file has not been opened yet, then we need to create it
		 * now.
		 */
		if (strcmp(tag, TAG_STREAM) == 0) {
			size_t len = strlen(params) >> 1;
			unsigned char *value = calloc(1, len);
			DecodeBlock(value, (unsigned char *)params);
			request_put_stream(tag, value, len);
			continue;
		}

		/* the encpass tag is bf_encrypt'd so the password
		 * will not have to be sent in plaintext.
		 */
		if (strcmp(tag, TAG_ENC_PASS) == 0) {
			char *password = bf_decrypt(params);
			request_del(tag);
			request_put(tag, password);
			continue;
		}

		if (strcmp(tag, TAG_ENV) == 0) {
			/* Discard any settings we don't actually understand. */
			if (*params == '\0' || NULL == strchr(SUPPORTED_ENV_ACTIONS, *params)) {
				if (bf_debug) {
					format_message(bf_debug, "Discarding unrecognized env type: %c", *params);
					DebugBlock(bf_debug);
				}
				continue;
			}
		}

		/* Check for a few invalid tags that would come from
		 * port scanners that think we speak HTTP.
		 */
		if (0 == strcmp(tag, "get") ||
				0 == strcmp(tag, "post") ||
				0 == strcmp(tag, "help") ||
				0 == strcmp(tag, "options"))
		{
			/* The telnet Are-You-There string will work fine for this. */
			io_ayt();
			break;
		}

		/* For anything else, we just save the value.  If we
		 * have already seen this tag, then append to it with
		 * this content, using a newline to separate it.  But
		 * first, make sure it is just text.  If somebody is
		 * sending us binary garbage, just hang up on them.
		 */
		p = tag;
		while ('\0' != *p && isalpha((int)(*p))) { ++p; }
		if ('\0' != *p) {
			io_ayt();
			break;
		}

		request_put(tag, params);
	}

	request_destroy();
	bf_strtok(NULL, NULL, NULL, &bf_sptr);
	free(buf);
	if (bf_debug) { DebugBlock("Quitting"); }
	return;
}



/*  
 *  process client requests
 */
int bfagent(void) {
	char    *buf;

	did_hello = 0;

	if (NULL != (buf = bfconf_get("debug_pause", NULL))) {
		platform_sleep(atoi(buf));
		free(buf);
	}

	/* Initialize locale and lang settings */
	{
		char *default_locale = bfconf_get("locale", NULL);
		if (NULL != default_locale) {
			platform_setenv("LANG", default_locale, 1);
			free(default_locale);
		}
	}
	platform_setlocale("");
	bfSetLang("");

	disable_telnet_support = BFCONF_EXISTS(disable_telnet_support);

	if (NULL != (buf = bfconf_get("heartbeat_interval", NULL))) {
		heartbeat_interval = atoi(buf);
		free(buf);
	}

	if (NULL != (buf = bfconf_get("command_output_cache", NULL))) {
		datablock_cache_size = atoi(buf);
		if (datablock_cache_size < 2048)
			datablock_cache_size = 2048;
		datablock_cache = malloc(datablock_cache_size + 1);
		free(buf);
	}

#ifdef WIN32
	memset(agent_mapped_drives, 0, sizeof(int) * 26);
#endif

	defMapStr = bfconf_get("map", NULL);
	request_new();


	if (agent_re_exec) {
		return platform_reexec_client();
	} else if (start_SSL() != NULL || agent_mode == SE_OK) {
		bfagent_main_loop();
		stop_SSL();
	}

	/* release memory we don't need. */
	platform_release_credentials();
	free(buf);
	request_destroy();
	platform_freeenv(clearcase_view);
	return 0;
}

