/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#ifndef BF_AGENT_H
#define BF_AGENT_H

/*
 * IBM Rational Build Forge Client Common Includes 
 */

/* Results from autoconf ... */
#include "config.h"

/* Do not trust setenv or unsetenv unless we have both of them. */
#ifdef HAVE_SETENV
#  ifndef HAVE_UNSETENV
#    undef HAVE_SETENV
#  endif
#else
#  undef HAVE_UNSETENV
#endif

/* Build information */
#define AGENT_ID_VERSION    "7.0.2.0-2-0065"

/* A few constants */
#define MAX_REQUEST_TAGS    32
#define MAX_TAG_LEN         28
#define BFBUFSIZ            16384
#define BFXFERBLOCKSIZE     8192
#define XFERLINELEN         128
#define MEM_REPORT_SIZE     1048576
#define DISK_REPORT_SIZE    1048576

/*
 * Windows includes
 */
#ifdef WIN32
/* #define _UNICODE */
#pragma warning (disable: 4706)
#define _WIN32_WINNT 0x0500
#include <winsock2.h>
#include <windows.h>
#include <ws2tcpip.h>
#include <io.h>
#include <process.h>
#include <sys/utime.h>
#define read(fd,buf,len) _read(fd,buf,len)
#define close(fd) _close(fd)

/*
 * UNIX/Linux/MacOS includes
 */
#else /* !def WIN32 */
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <poll.h>
#include <sys/socket.h>
#include <sys/wait.h>
#endif /* WIN32 */


/*
 * Required includes for all platforms
 */
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>



/*
 * Optional includes
 */

#ifdef HAVE_CTYPE_H
#  include <ctype.h>
#endif
#ifdef HAVE_GETOPT_H
#  include <getopt.h>
#endif
#ifdef HAVE_ICONV_H
#  include <iconv.h>
#endif
#ifdef HAVE_IFADDRS_H
#  include <ifaddrs.h>
#endif
#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif
#ifdef HAVE_LANGINFO_H
#  include <langinfo.h>
#endif
#ifdef HAVE_LIMITS_H
#  include <limits.h>
#endif
#ifdef HAVE_LOCALE_H
#  include <locale.h>
#endif
#ifdef HAVE_MACH_MACH_H
#  include <mach/mach.h>
#endif
#ifdef HAVE_MALLOC_H
#  include <malloc.h>
#endif
#ifdef HAVE_NETDB_H
#  include <netdb.h>
#endif
#ifdef HAVE_NETINET_IN_H
#  include <netinet/in.h>
#endif
#ifdef HAVE_NET_IF_H
#  if defined(ZOS) && !defined(_OE_SOCKETS)
#    define _OE_SOCKETS
#  endif
#  include <net/if.h>
#  undef _OE_SOCKETS
#endif
#ifdef HAVE_NET_IF_DL_H
#  include <net/if_dl.h>
#endif
#ifdef HAVE_NET_IF_PACKET_H
#  include <net/if_packet.h>
#endif
#ifdef HAVE_NETPACKET_PACKET_H
#  include <netpacket/packet.h>
#endif
#ifdef HAVE_REGCOMP
#  include <regex.h>
#endif
#ifdef HAVE_SYS_SYSCTL_H
#  include <sys/sysctl.h>
#endif
#ifdef HAVE_SYS_VMMETER_H
#  include <sys/vmmeter.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#  include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_LOADAVG_H
#  include <sys/loadavg.h>
#endif
#ifdef HAVE_SYS_MOUNT_H
#  include <sys/mount.h>
#endif
#ifdef __BSD_VISIBLE
#  include <sys/param.h>
#endif
#ifdef HAVE_SYS_SOCKIO_H
#  include <sys/sockio.h>
#endif
#ifdef HAVE_SYS_STATFS_H
#  include <sys/statfs.h>
#endif
#ifdef HAVE_SYS_STATVFS_H
#  include <sys/statvfs.h>
#endif
#ifdef HAVE_SYS_UTSNAME_H
#  include <sys/utsname.h>
#endif
#ifdef HAVE_SYS_VMINFO_H
#  include <sys/vminfo.h>
#endif
#ifdef HAVE_TERMIOS_H
#  include <termios.h>
#endif
#ifdef HAVE_TERMIO_H
#  include <termio.h>
#endif
#ifdef HAVE_UTIME_H
#  include <utime.h>
#endif
#ifdef HAVE_UTIME_H
#  include <utime.h>
#endif
#ifdef HAVE_ARPA_INET_H
#  include <arpa/inet.h>
#endif
#ifdef HAVE_SSL
#  define OPENSSL_NO_KRB5
#  include <bio.h>
#  include <err.h>
#  include <rand.h>
#  include <ssl.h>
#  include <x509v3.h>
#endif

#ifdef ISERIES_AGENT
#  include "as400_protos.h"
#endif


/* Fix gcc on HP/UX 11 */
#if defined(__GNUC__) && !defined(_BSIZE_T)
#  define _BSIZE_T
#  if defined(_APP32_64BIT_OFF_T) || defined(_KERNEL)
	typedef int64_t sbsize_t;
	typedef uint64_t bsize_t;
#  else
	typedef long sbsize_t;
	typedef unsigned long bsize_t;
#  endif
#endif /* __GNUC__ && !_BSIZE_T */


/*
 * implementations for potentially missing functions, data structures,
 * data structure members, constants, etc.
 */
#ifndef HAVE_ICONV_H
typedef int iconv_t;
#  define iconv_open(tocode,fromcode) ((iconv_t)-1)
#  define iconv(cd,inbuf,inbytesleft,outbuf,outbytesleft) (errno = EBDAF, ((size_t)-1))
#  define iconv_close(cd) (errno = EBADF, ((size_t)-1))
#endif /* !HAVE_ICONV_H */

#ifndef ICONV_INVALID
#  define ICONV_INVALID ((iconv_t)-1)
#endif /* !ICONV_INVALID */

#ifndef HAVE_MEMRCHR
#  undef memrchr
#  define memrchr(s,c,n) bf_memrchr(s,c,n)
void *bf_memrchr(const void *s, int c, size_t n);
#endif /* !HAVE_MEMRCHR */

#ifndef HAVE_REGCOMP
#  include "bf_regex.h"
#endif

#ifndef HAVE_STRDUP
#  undef strdup
#  define strdup(x) BFStrDup(x)
char *BFStrDup(const char *);
#endif /* !HAVE_STRDUP */

#ifdef HAVE_SOCKADDR_SA_LEN
#  define BF_SA_LEN(saddr) ( (saddr)->sa_len > sizeof(*(saddr)) \
	 	? (saddr)->sa_len \
	 	: sizeof(*(saddr)))
# else /* !HAVE_SOCKADDR_SA_LEN */
#  define BF_SA_LEN(saddr) (sizeof(*(saddr)))
#endif /* HAVE_SOCKADDR_SA_LEN/! */

#ifndef O_BINARY
#  define O_BINARY (0)
#endif

#ifndef SEEK_SET
#  define	SEEK_SET (0)
#endif

#ifndef S_IRUSR
#  define S_IRUSR (0200)
#endif

#ifndef S_IWUSR
#  define S_IWUSR (0400)
#endif

#ifndef CODESET
#  define CODESET (0)
#  undef nl_langinfo
#  define nl_langinfo(item) ("ASCII")
#endif

#ifndef HAVE_LOCALE_H
#  undef setlocale
#  define setlocale(category,locale) (NULL)
#endif

#ifndef MAX_PATH
#  ifdef PATH_MAX
#    define MAX_PATH PATH_MAX
#  elif defined(MAXPATHLEN)
#    define MAX_PATH MAXPATHLEN
#  else
#    define MAX_PATH BFBUFSIZ
#  endif
#endif /* !def MAX_PATH */




/*
 * Override memory calls
 */
#if defined(DEBUG_MEM) && !defined(MEM_C)
void *bf_malloc(size_t size);
void *bf_calloc(size_t nmemb, size_t size);
void *bf_realloc(void *ptr, size_t size);
void bf_free(void *ptr);
#define malloc(size) bf_malloc(size)
#define calloc(nmemb,size) bf_calloc(nmemb,size)
#define free(ptr) bf_free(ptr)
#define realloc(ptr,size) bf_realloc(ptr,size)
#endif




/*======================================================================
 * And finally, the real Build Forge stuff...
 *======================================================================*/


#define BFDEBUG(x)	if (bf_debug) DebugBlock(x);

#define BOURNE_REDIRECT	"%s > %s 2>&1"
#define	CSH_REDIRECT "%s >& %s"


/*
 * dot commands that aren't internal but are translated/located
 */
#define CMD_MAKE        1

#define SE_OK           0
#define SE_OK_SECURE    1
#define SE_OK_CERTIFIED 2
#define SE_INIT         3
#define SE_BIOIN        4
#define SE_BIOOUT       5
#define SE_BADCA        6
#define SE_BADCERT      7
#define SE_BADPKEYFILE  8
#define SE_BADPKEY      9
#define SE_NOSSL        10
#define SE_HANDSHAKE    11

/*
 * config entries
 */
#define CONFIG_CERT     "cert"
#define CONFIG_CA       "ca"

/*
 *  well known request tags
 */
#define TAG_APPEND      "append"
#define TAG_CHECKPOINT  "checkpoint"
#define TAG_CLOBBER     "clobber"
#define TAG_COMMAND     "cmd"
#define TAG_DATA        "data"
#define TAG_DEBUG       "debug"
#define TAG_DIR         "dir"
#define TAG_ENC_PASS    "encpass"
#define TAG_ENV         "env"
#define TAG_EXEC        "exec"
#define TAG_FLAGS       "flags"
#define TAG_LANG        "lang"
#define TAG_LICENSE     "key"
#define TAG_MODE        "mode"
#define TAG_OFFSET      "offset"
#define TAG_PASSWORD    "password"
#define TAG_PATH        "path"
#define TAG_QUIT        "quit"
#define TAG_RECURSIVE   "recursive"
#define TAG_ROOT        "root"
#define TAG_SILENT      "silent"
#define TAG_STAMP       "stamp"
#define TAG_STDIN       "stdin"
#define TAG_STREAM      "stream"
#define TAG_TAG         "tag"
#define TAG_TIMEOUT     "timeout"
#define TAG_TRIGGER     "go"
#define TAG_USERDATA    "userdata"
#define TAG_USERNAME    "username"



/*
 *  result codes and messages 
 */

/*
 *  200's
 */
#define CODE_HELLO              200
#define CODE_SECURE_HELLO       201
#define CODE_CERTIFIED_HELLO	202
#define CODE_GOODBYE            210
#define END_MESSAGE             "260 EOR\n"

/*
 *  300's
 */
#define CODE_DATA           300
#define CODE_DATA_MSG       "DATA"

#define SU_OK               0
#define SU_UNKNOWNUSER      1
#define SU_SETUSERFAILED    2
#define SU_PAMSESSIONERROR  3
#define SU_ILLEGAL          4
#define SU_OK_NOACCOUNT     5
#define SU_USERNOTFOUND     6
#define SU_ACCOUNTINVALID   7
#define SU_NOCRED           8
#define SU_NOSESSION        9

/* Subsystem IDs */
#define SUBS_EXEC           "EXEC"
#define SUBS_MKDIR          "MKDIR"
#define SUBS_RMDIR          "RMDIR"
#define SUBS_COPY           "COPY"
#define SUBS_READ           "READ"
#define SUBS_WRITE          "WRITE"
#define SUBS_SPACE          "SPACE"


/*
 *  result codes
 */
#define RESULT_OK           220
#define RESULT_UNKNOWN      250





typedef enum {
	CACHE_ON,
	CACHE_OFF,
	CACHE_FLUSH
} cache_request_t;




#ifndef WIN32
extern char **environ;
#endif
extern int verbose;
extern char *hiddenEnv;
extern int env_recursion_limit;
extern char *bf_debug;
extern int agent_mode;
extern char *mapStr;
extern char *defMapStr;
extern char *trapCommand;
extern int trapmode;
extern char *clearcase_view;
extern int cc_audit;
extern char *LANG;
extern void bfSetLang(const char *);
const char *bfCat(int);
extern char *conf_warning;
extern int datablock_cache_size;
extern char *datablock_cache;
extern int heartbeat_interval;
extern char **native_extensions;

/*
 *  some commonly used functions 
 */
void cleanMyCruft(char *);
int format_message(char *, const char *, ...);
int io_ayt(void);
int     io_puts(const char *);
char *  io_gets(char *, int);
int     checkIP(const char *);
const char *bf_langinfo_codeset(void);
#define	MAX_REGEX_SUBSTRINGS	10
regex_t *bf_regex_compile(char *,int);
void    bf_regex_free(regex_t *);
/* char 	*bf_regex_replace(char *,char *,char *); */
int     bf_regex_prep(regex_t **, char **, char **, char **, char *);
char    *bf_regex_apply(const char *, const regex_t *, const char *);
char    *bf_strtok(char *, const char *, char **, char **);
char	*bf_strpart(char *, char *);
char	**getArgs(char *);
char	**getArgsN(char *, int);
char	**getArgsC(char *, char);
void	freeArgs(char **);
char	*shiftArgs(char **);
int 	ResultBlock(int, int, const char *);
int 	DataBlock(const char *,const  char *);
int 	DebugBlock(const char *);
void 	MonitorOpen(void);
void 	MonitorClose(void);
void 	MonitorOutput(const char *);
int 	HeartBeat(void);
int 	OutputCache(cache_request_t);
char	*LowerString(char *);
char	*toForwardSlashes(char *);
char	*toBackSlashes(char *);
char	*CollapseSlashes(char *);
char	*CleanUnixPath(char *);
char	*CleanWindowsPath(char *);
char	*CleanVOSPath(char *);
char	*RemoveNewline(char *);
char	*RemoveTrailingNewline(char *);
int 	BFCreateDirectory(char *, char *, int);
int 	BFCreateDirectoryRelative(char *, int);
void 	BFRemoveDirectory(char *);
int 	isDir(const char *);
int 	isFile(const char *);
void	DumpEnv(void);
void	BuildEnv(char *,char **);
char	*CleanPath(char *);
char	*HasEOL(char *);
char	**getBFenviron(void);
void	freeBFenviron(char **);
char	*bf_decrypt(const char *);
int 	isHidden(const char *);
void 	addHidden(char *);
void 	freeHidden(void);
char	*EncodeBlock(const unsigned char *, int, int);
int 	DecodeBlock(unsigned char *, const unsigned char *);
void	OpenLog(void);
void	CloseLog(void);
char	EnvPathSeparator(void);
char	PathSeparator(void);
char	*this_view(void);
int 	BFStrCmp(const char *, const char *, int);
int 	SetUserAccount(void);
int 	UnSetUserAccount(void);
int		check_magic_login(const char *, const char *);
int 	BecomeUser(const char *, const char *);
int 	UnBecomeUser(void);
int 	CCDiffView(char *, struct stat *, int);
void	sayHello(int);
void 	*start_SSL(void);
void	stop_SSL(void);
char	*formatDate(char *, struct tm *);
char	*formatDateStamp(char *, time_t);
char	*formatUTCDateStamp(char *, time_t);
char	*expandDate(char *);
char	*expandGMDate(char *);
#ifdef WIN32
char	*strtok_r(char *, const char *, char **);
int     getopt(int, char * const *, const char *);
extern int cygwin;
extern int agent_mapped_drives[];
#endif
void 	ClobberPasswords(void);
int 	valid_credentials(const char *,const char *,const char *);
int 	bf_LoadValue(void);
int 	bf_dynamic_function(char *, int , int *, int *);
void	valid_credentials_release(void);
char	*clean_win_slashes(char *);
char	*clean_win_leading_slashes(char *);

uint32_t get_crc_string(char *);
uint32_t get_crc_file(char *);
void    gen_crc_table(void);
void    StoreUserData(char *, size_t);
void    FreeUserData(void);
int     run_process(char *, char *, int , char *, char *, char *, int *, int);
char *  CCStartViews(int);
int     CCMountVobs(int);


#ifdef HAVE_SSL
void	report_SSL(void);
#endif


#ifdef WIN32
#  undef inline
#  define inline __inline
#  undef HAVE_X86_ASM
#  define HAVE_X86_ASM 2
#endif

#include "bf_funcs.h"
#include "bf_ipv6.h"
#include "bfconf.h"
#include "codes.h"
#include "convert.h"
#include "env.h"
#include "hash.h"
#include "platform.h"
#include "request.h"


#endif /* BF_AGENT_H */

