/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

bf_internal_t internals[] = {
	{ ".strsub",    &builtinStrsub },
	{ ".edit",      &builtinEdit },
	{ ".mkdir",     &builtinMkdir },
	{ ".rmdir",     &builtinRmdir },
	{ ".require",   &builtinRequire },
	{ ".monitor",   &builtinMonitor },
	{ ".alive",     &builtinAlive },
	{ NULL, NULL }
};

int builtinMkdir(char *paths, int timeout, int *timed_out)
{
	char	**pathlist;
	int 	i, len;
	int		ret = 0;
	char    *buf = malloc(BFBUFSIZ);
	char    *msg = malloc(BFBUFSIZ);
	char    *root = malloc(MAX_PATH);

	pathlist = getArgs(paths);
	platform_getcwd(root, MAX_PATH);

	for (i=0; pathlist && pathlist[i]; ++i) {
		CleanUnixPath(pathlist[i]);
		len = strlen(root);
		memcpy(buf, root, len);
		buf[len++] = '/';
		strcpy(&buf[len], pathlist[i]);
		CleanUnixPath(buf);
		if (isDir(buf)) {
			CleanPath(buf);
			format_message(msg, bfCat(E_MKDIR_EXISTS), buf);
			DataBlock(msg, NULL);
			continue;
		}
		CleanPath(buf);
		if (BFCreateDirectoryRelative(pathlist[i], 0) != 0) {
			format_message(msg, bfCat(E_MKDIR_FAIL), buf);
			DataBlock(msg, NULL);
			ret = E_MKDIR_FAIL;
			break;
		}
		format_message(msg, bfCat(E_MKDIR_OK), buf);
		DataBlock(msg, NULL);
	}

	freeArgs(pathlist);
	free(root);
	free(msg);
	free(buf);
	return ret;
}

int builtinRmdir(char *paths, int timeout, int *timed_out)
{
	char	*dir, **pathlist;
	int 	i, len, ret = 0;
	char	*buf = malloc(BFBUFSIZ);
	char	*msg = malloc(BFBUFSIZ);
	char    *root = malloc(MAX_PATH);

	pathlist = getArgs(paths);
	platform_getcwd(root, MAX_PATH);
	msg = malloc(BFBUFSIZ);

	for (i = 0 ; pathlist && pathlist[i] ; i++) {
		dir = pathlist[i];
		len = strlen(root);
		memcpy(buf, root, len);
		buf[len++] = '/';
		strcpy(&buf[len], dir);
		CleanPath(buf);

		if (0 != platform_access(buf, 0)) {
			switch (errno) {
				case EACCES:
					format_message(msg, bfCat(E_RMDIR_NOTTHERE_EACCESS), buf);
					break;
				case ENOENT:
					format_message(msg, bfCat(E_RMDIR_NOTTHERE_ENOENT), buf);
					break;
				case ENOTDIR:
					format_message(msg, bfCat(E_RMDIR_NOTTHERE_ENOTDIR), buf);
					break;
				default:
					format_message(msg, bfCat(E_RMDIR_NOTTHERE), buf);
					break;
			}
			DataBlock(msg, NULL);
		}
		else {
			BFRemoveDirectory(buf);
			if (0 == platform_access(buf, 0)) {
				format_message(msg, bfCat(E_RMDIR_FAIL), buf);
				DataBlock(msg, NULL);
				ret = E_RMDIR_FAIL;
				break;
			}
			format_message(msg, bfCat(E_RMDIR_OK), buf);
			DataBlock(msg, NULL);
		}
	}

	freeArgs(pathlist);
	free(root);
	free(msg);
	free(buf);
	return ret;
}

static int require_disk(char *buf, char *req)
{
	const char *path = &req[4]; /* disk**** */
	char *p;
	platform_sysconf_disk_t *disk;
	int rc = 0;
	long have;
	long need;

	if (*path == '(') {
		/* disk(/path/name)=**** */
		path++;
		if (NULL == ( p = strchr(path, ')') )) {
			format_message(buf, bfCat(E_REQUIRE_DISK_MALFORMED), req);
			DataBlock(buf, NULL);
			return 3;
		}
		*p++ = 0;
		if (NULL != ( p = strchr(p, '=') ))
			*p++ = 0;
		need = atol(p);
	} else {
		/* disk=**** */
		need = atol(++path);
		path = ".";
	}

	if (NULL == ( disk = platform_sysconf_disk_new(path) )) {
		DataBlock(bfCat(E_REQUIRE_DISK_UNSUPP), NULL);
		return 0;
	}
	have = (long)(disk->disk_free / DISK_REPORT_SIZE);
	platform_sysconf_disk_destroy(disk);

	if (have < 0) {
		rc = 2;
		format_message(buf, bfCat(E_REQUIRE_DISK_BAD_PATH), path);
	} else if (have < need) {
		rc = 1;
		format_message(buf, bfCat(E_REQUIRE_DISK_INSUFF), path, have, need);
	} else {
		format_message(buf, bfCat(M_REQUIRE_DISK_OK), path, have, need);
	}

	DataBlock(buf, NULL);
	return rc;
}

static int require_mem(char *buf, char *req)
{
	platform_sysconf_mem_t *mem;
	int rc = 0;
	long need = atol(&req[4]); /* mem=**** */
	long have;

	if (NULL == ( mem = platform_sysconf_mem_new() )) {
		DataBlock(bfCat(E_REQUIRE_MEM_UNSUPP), NULL);
		return 0;
	}
	have = (long)(mem->mem_total / MEM_REPORT_SIZE);
	platform_sysconf_mem_destroy(mem);

	if (have < need) {
		rc = 1;
		format_message(buf, bfCat(E_REQUIRE_MEM_INSUFF), have, need);
	} else {
		format_message(buf, bfCat(M_REQUIRE_MEM_OK), have, need);
	}

	DataBlock(buf, NULL);
	return rc;
}

int builtinRequire(char *cl, int timeout, int *timed_out)
{
	char ** argv;
	char *  buf;
	int     argc;
	int     result;
	int     negate;

	result = 0;
	negate = 0;
	buf = calloc(1, BFBUFSIZ);
	argv = getArgs(cl);
	for (argc=0; argv && argv[argc]; ++argc) {
		if (strncmp(argv[argc], "-n", 2) == 0) {
			negate = 1;
			DataBlock(bfCat(M_REQUIRE_NEGATING), NULL);
			continue;
		}

		if (strncmp(argv[argc], "disk", 4) == 0) {
			int rc = require_disk(buf, argv[argc]);
			if (0 != rc)
				result = rc;
			continue;
		}
		
		if (strncmp(argv[argc], "mem=", 4) == 0) {
			int rc = require_mem(buf, argv[argc]);
			if (0 != rc)
				result = rc;
			continue;
		}

		format_message(buf, bfCat(E_REQUIRE_HUH), argv[argc]);
		DataBlock(buf, NULL);
		freeArgs(argv);
		free(buf);
		return -1;
	}
	freeArgs(argv);
	free(buf);
	return negate ? !result : result;
}

int builtinMonitor(char *cl, int timeout, int *timed_out)
{
	char	**argv, *path;
	int 	argc, delay, waitforit, docat;
	struct stat sb;
	size_t	lastsize;
	char	*msg;
	time_t	now, expires;
	FILE	*f;

	msg = calloc(1, BFBUFSIZ);
	argv = getArgs(cl);
	waitforit = docat = 0;
	for (argc = 0 ; argv && argv[argc] ; argc++) {
		if (strcmp(argv[argc], "-w") == 0) {
			waitforit = 1;
			free(shiftArgs(argv));
			argc--;
		}
		else if (strcmp(argv[argc], "-c") == 0) {
			docat = 1;
			free(shiftArgs(argv));
			argc--;
		}
	}

	if (argc < 2) {
		DataBlock(bfCat(E_MONITOR_USAGE), NULL);
		free(msg);
		freeArgs(argv);
		return -1;
	}
	delay = atoi(argv[0]);
	path = argv[1];

	if (delay < 1) {
		DataBlock(bfCat(E_MONITOR_USAGE), NULL);
		free(msg);
		freeArgs(argv);
		return -1;
	}

	if (waitforit) {
		time(&now);
		expires = now + timeout;
		format_message(msg, bfCat(E_MONITOR_WAITINGFOR), path);
		DataBlock(msg, NULL);
		while (0 != platform_stat(path, &sb)) {
			time(&now);
			if (timeout && now > expires) {
				if (timed_out)
					*timed_out = timeout;
				format_message(msg, bfCat(E_MONITOR_TIMEOUT), path);
				DataBlock(msg, NULL);
				free(msg);
				freeArgs(argv);
				return -1;
			}
			platform_sleep(delay);
		}
	}

	lastsize = -1;
	if (0 != platform_stat(path, &sb)) {
		format_message(msg, bfCat(E_MONITOR_STATFAIL), path);
		DataBlock(msg, NULL);
		free(msg);
		freeArgs(argv);
		return -1;
	}
	while ((unsigned long)lastsize != (unsigned long)sb.st_size) {
		lastsize = sb.st_size;
		format_message(msg, bfCat(E_MONITOR_SIZEOK), path, (long)lastsize);
		DataBlock(msg, NULL);
		platform_sleep(delay);
		platform_stat(path, &sb);
	}
	format_message(msg, bfCat(E_MONITOR_STOPPED), path, (long)lastsize, delay);
	DataBlock(msg, NULL);

	if (docat) {
		f = platform_fopen(path, "r");
		if (f != NULL) {
			while (platform_fgets(msg, BFBUFSIZ, f) != NULL) {
				RemoveNewline(msg);
				DataBlock(msg, NULL);
			}
			fclose(f);
		}
	}

	free(msg);
	freeArgs(argv);
	return 0;
}

static int attempt_connect(struct addrinfo *ai_head, int poll_timeout) {
	struct addrinfo *ai;
	int rc = E_ALIVE_BADHOST;

	for (ai=ai_head; NULL != ai; ai=ai->ai_next) {
		int sock = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (sock == -1) {
			rc = E_ALIVE_NOSOCKET;
		} else {
			platform_alarm(poll_timeout);
			rc = connect(sock, ai->ai_addr, ai->ai_addrlen);
			platform_alarm(0);
			platform_closesocket(sock);
			if (0 == rc) { break; }
			rc = E_ALIVE_FAILED;
		}
	}
	return rc;
}

/*
 * Poll a remote location for connections.
 */
static int polling_connect(
		const char *target_hostname,
		const char *target_port,
		char *buf,
		int poll_timeout,
		int master_timeout)
{
	time_t  deadline = time(NULL) + master_timeout;
	int     rc = 0;
	int     count = 0;
	struct  addrinfo hints;
	struct  addrinfo *ai;

	/* resolve hostname and port */
	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	if (0 != getaddrinfo(target_hostname, target_port, &hints, &ai)) {
		format_message(buf, bfCat(E_ALIVE_BADHOST));
		DataBlock(buf, NULL);
		free(buf);
		return E_ALIVE_BADHOST;
	}

	platform_signals_catch("bfagent");
	while (time(NULL) <= deadline) {
		format_message(buf, bfCat(E_ALIVE_ATTEMPT),
				target_hostname, target_port, ++count);
		DataBlock(buf, NULL);

		rc = attempt_connect(ai, poll_timeout);

		if (rc == 0) { break; }
		platform_sleep(poll_timeout);
	}
	freeaddrinfo(ai);

	switch (rc) {
		case 0:
			format_message(buf, bfCat(E_ALIVE_OK), target_hostname, target_port);
			break;
		case E_ALIVE_BADHOST:
			format_message(buf, bfCat(E_ALIVE_BADHOST));
			break;
		case E_ALIVE_NOSOCKET:
			format_message(buf, bfCat(E_ALIVE_NOSOCKET), errno);
			break;
		default:
			rc = E_ALIVE_FAILED;
			format_message(buf, bfCat(rc), target_hostname, target_port);
			break;
	}
	platform_signals_release();

	DataBlock(buf, NULL);
	free(buf);
	return rc;
}

int builtinAlive(char *cl, int master_timeout, int *timed_out)
{
	char ** argv;
	char *  target_hostname;
	char *  target_port;
	char *  buf = calloc(1, BFBUFSIZ);
	int 	rc = E_ALIVE_USAGE;
	int     poll_timeout = 0;

	/* Parse command line args */
	argv = getArgs(cl);

	/* set the overall timeout to default 5m or the passed in value.
	 * unlike most steps, the user is not permitted to specify an
	 * infinite timeout value for .alive commands, because then there
	 * would be no way for .alive to ever fail.  It would either
	 * connect successfully or never return.  It's important to know
	 * when to give up.
	 */
	if (master_timeout == 0) { master_timeout = 300; }
	*timed_out = 0;

	/* parse/set the args: hostname port poll_timeout */
	if (!argv || !*argv) { goto done; }
	target_hostname = *argv++;
	if (!*argv || '\0' == target_hostname) { goto done; }
	target_port = *argv++;
	if (!*argv || '\0' == target_port) { goto done; }
	poll_timeout = atoi(*argv);
	if (poll_timeout <= 0) { goto done; }

	rc = polling_connect(target_hostname, target_port,
			buf, poll_timeout, master_timeout);

done:
	if (rc == E_ALIVE_USAGE) {
		format_message(buf, bfCat(E_ALIVE_USAGE));
		DataBlock(buf, NULL);
	}
	freeArgs(argv);
	free(buf);
	return rc;
}

/*
 * args:  /srch/repl/ pathname
 *
 */
int builtinEdit(char *arg, int timeout, int *timed_out) {
	FILE	*inf, *outf;
	regex_t	*srch;
	char	*path, *tmpname, *repl, *target;
	char	**args, *sav, *newbuf, *spat, *cl;
	char	*buf = malloc(BFBUFSIZ);
	int 	i, lines, changes;
	int 	transcode = platform_utf8_is_native() ? 0 : 1;
	char	transcode_buf[BFBUFSIZ];

	buf = calloc(sizeof(*buf), BFBUFSIZ);
	cl = strdup(arg);
	args = getArgs(arg);

	if (!args || !args[0] || !args[1]) {
		DataBlock(bfCat(E_EDIT_WRONG_ARGS), NULL);
		freeArgs(args);
		free(buf);
		free(cl);
		return -1;
	}
	freeArgs(args);
	srch = NULL;
	repl = target = spat = NULL;
	if (bf_regex_prep(&srch, &spat, &repl, &target, cl) != 0 || NULL == srch) {
		format_message(buf, ".edit: %s", target);
		DataBlock(buf, NULL);
		free(target);
		free(buf);
		return -1;
	}

	args = getArgs(target);
	if (!args || !args[0]) {
		DataBlock(bfCat(E_EDIT_BAD_EXPRESSION), NULL);
		free(spat);
		free(target);
		if (repl)
			free(repl);
		free(buf);
		bf_regex_free(srch);
		freeArgs(args);
		return -1;
	}

	for (i = 0 ; args[i] ; i++) {
		path = args[i];
		CleanPath(path);
		format_message(buf, bfCat(M_EDIT_APPLYING), spat, *spat, repl, *spat, path);
		DataBlock(buf, NULL);
		inf = platform_fopen(path, "rb");

		if (inf == NULL) {
			format_message(buf, bfCat(E_CONFOPENFAIL), path, errno);
			DataBlock(buf, NULL);
			free(buf);
			freeArgs(args);
			if (repl)
				free(repl);
			if (target)
				free(target);
			free(spat);
			if (srch)
				bf_regex_free(srch);
			return 2;
		}
	
		/* make a tmp file copy */
		tmpname = platform_tempnam();
		outf = platform_fopen(tmpname, "wb");
		if (outf == NULL) {
			format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, tmpname);
			DataBlock(buf, NULL);
			free(buf);
			free(tmpname);
			freeArgs(args);
			if (repl)
				free(repl);
			if (target)
				free(target);
			free(spat);
			bf_regex_free(srch);
			return 3;
		}
	
		/* copy the contents */
		errno = 0;
		while (fgets(buf, BFBUFSIZ, inf) != NULL) {
			if (fputs(buf, outf) == EOF) {
				fclose(inf);
				fclose(outf);
				format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, tmpname);
				DataBlock(buf, NULL);
				free(tmpname);
				free(buf);
				freeArgs(args);
				if (repl)
					free(repl);
				if (target)
					free(target);
				free(spat);
				bf_regex_free(srch);
				return 4;
			}
		}
		if (!feof(inf) && errno) {
			fclose(inf);
			fclose(outf);
			format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, tmpname);
			DataBlock(buf, NULL);
			free(tmpname);
			free(buf);
			freeArgs(args);
			if (repl)
				free(repl);
			if (target)
				free(target);
			bf_regex_free(srch);
			free(spat);
			return 5;
		}
	
		/* close the handles */
		fclose(inf);
		fclose(outf);
	
		/* now do the replacement */
		inf = platform_fopen(tmpname, "rb");
		if (inf == NULL) {
			format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, tmpname);
			DataBlock(buf, NULL);
			free(buf);
			platform_unlink(tmpname);
			free(tmpname);
			freeArgs(args);
			if (repl)
				free(repl);
			if (target)
				free(target);
			bf_regex_free(srch);
			free(spat);
			return 6;
		}

		outf = platform_fopen(path, "wb");
		if (outf == NULL) {
			format_message(buf, bfCat(E_WRITE_FAILORIG), path, errno, tmpname);
			DataBlock(buf, NULL);
			free(buf);
			free(tmpname);
			freeArgs(args);
			if (repl)
				free(repl);
			if (target)
				free(target);
			bf_regex_free(srch);
			free(spat);
			return 7;
		}
	
		/* copy the contents, doing substitution */
		errno = 0;
		lines = changes = 0;
		while (fgets(buf, BFBUFSIZ >> 1, inf) != NULL) {
			char *outbuf;

			++lines;
			if (transcode) {
				convert_native_to_utf8(buf, -1, transcode_buf, BFBUFSIZ);
				transcode_buf[BFBUFSIZ-1] = '\0';
				sav = strdup(transcode_buf);
			} else {
				sav = strdup(buf);
			}
			newbuf = bf_regex_apply(sav, srch, repl);
			if (strcmp(sav, newbuf)) {
				++changes;
				format_message(buf, bfCat(M_EDIT_BEFORE), lines, sav);
				RemoveNewline(buf);
				DataBlock(buf, NULL);
				format_message(buf, bfCat(M_EDIT_AFTER), newbuf);
				RemoveNewline(buf);
				DataBlock(buf, NULL);
			}
			if (transcode) {
				convert_utf8_to_native(newbuf, -1, transcode_buf, BFBUFSIZ);
				outbuf = transcode_buf;
			} else {
				outbuf = newbuf;
			}
			if (fputs(outbuf, outf) == EOF) {
				fclose(inf);
				fclose(outf);
				format_message(buf, bfCat(E_WRITE_FAILORIG), path, errno, tmpname);
				DataBlock(buf, NULL);
				free(tmpname);
				free(buf);
				free(newbuf);
				freeArgs(args);
				if (repl)
					free(repl);
				if (target)
					free(target);
				bf_regex_free(srch);
				free(spat);
				return 8;
			}
			free(newbuf);
			free(sav);
		}
		if (!feof(inf) && errno) {
			fclose(inf);
			fclose(outf);
			format_message(buf, bfCat(E_READ_FAILORIG), errno, tmpname);
			DataBlock(buf, NULL);
			free(tmpname);
			free(buf);
			freeArgs(args);
			if (repl)
				free(repl);
			if (target)
				free(target);
			bf_regex_free(srch);
			free(spat);
			return 9;
		}
		fclose(inf);
		fclose(outf);
	
		format_message(buf, bfCat(M_EDIT_RESULTS), lines, changes);
		DataBlock(buf, NULL);
	
		/* clean up! */
		platform_unlink(tmpname);
		free(tmpname);
	}
	
	free(buf);
	freeArgs(args);
	if (repl)
		free(repl);
	if (target)
		free(target);
	bf_regex_free(srch);
	free(spat);
	return 0;
}


/*
 * args: search replace pathname
 * this was the original .strsub routine from 3.1 agents.
 *
 */
#define BF_STRSUB_BLOCK_SIZE (BFBUFSIZ >> 1)
int builtinStrsub(char *arg, int timeout, int *timed_out) {
	FILE	*inf, *outf;
	char	*buf, *tmpname, *msg, *newbuf, *q, *s, *utf8_buf;
	int 	i, lines, changes;
	char	*srch, *repl, *path;
	char	**args = getArgs(arg);
	size_t	srchlen, repllen;
	int 	transcode = platform_utf8_is_native() ? 0 : 1;
	char	transcode_buf[BFBUFSIZ];

	if (!args || !args[0] || !args[1] || !args[2]) {
		DataBlock(bfCat(E_STRSUB_HUH), NULL);
		freeArgs(args);
		return 1;
	}
	srch = args[0]; srchlen = strlen(srch);
	repl = args[1]; repllen = strlen(repl);

	for (i = 2 ; args[i] ; i++) {
		path = args[i];
		CleanPath(path);
		buf = calloc( sizeof(*buf),
				strlen(path) + srchlen + repllen + BF_STRSUB_BLOCK_SIZE );
	
		/* open the original file */
		inf = platform_fopen(path, "rb");
		if (inf == NULL) {
			format_message(buf, bfCat(E_STRSUB_OPEN_R), path);
			DataBlock(buf, NULL);
			free(buf);
			freeArgs(args);
			return 2;
		}
	
		/* make a tmp file copy */
		tmpname = platform_tempnam();
		outf = platform_fopen(tmpname, "wb");
		if (outf == NULL) {
			format_message(buf, bfCat(E_STRSUB_OPEN_W_TMP), tmpname);
			DataBlock(buf, NULL);
			free(buf);
			platform_unlink(tmpname);
			free(tmpname);
			freeArgs(args);
			return 3;
		}
	
		/* copy the contents */
		errno = 0;
		while (fgets(buf, BFBUFSIZ, inf) != NULL) {
			if (fputs(buf, outf) == EOF) {
				fclose(inf);
				fclose(outf);
				format_message(buf, bfCat(E_STRSUB_WRITE_ERROR), errno, tmpname);
				DataBlock(buf, NULL);
				free(tmpname);
				free(buf);
				freeArgs(args);
				return 4;
			}
		}
		if (!feof(inf) && errno) {
			fclose(inf);
			fclose(outf);
			format_message(buf, bfCat(E_STRSUB_READ_ERROR), errno, path);
			DataBlock(buf, NULL);
			free(tmpname);
			free(buf);
			freeArgs(args);
			return 5;
		}
	
		/* close the handles */
		fclose(inf);
		fclose(outf);
	
		/* now do the replacement */
		inf = platform_fopen(tmpname, "rb");
		if (inf == NULL) {
			format_message(buf, bfCat(E_STRSUB_OPEN_R_TMP), tmpname);
			DataBlock(buf, NULL);
			free(buf);
			free(tmpname);
			freeArgs(args);
			return 6;
		}

		outf = platform_fopen(path, "wb");
		if (outf == NULL) {
			format_message(buf, bfCat(E_STRSUB_OPEN_W), path, tmpname);
			DataBlock(buf, NULL);
			free(buf);
			free(tmpname);
			freeArgs(args);
			return 7;
		}
	
		/* copy the contents, doing substitution */
		errno = 0;
		format_message(buf, bfCat(M_STRSUB_REPLACING), srch, repl, path);
		DataBlock(buf, NULL);
		newbuf = calloc(1, BF_STRSUB_BLOCK_SIZE);
		lines = changes = 0;
		while (fgets(buf, BF_STRSUB_BLOCK_SIZE, inf) != NULL) {
			++lines;
			if (transcode) {
				convert_native_to_utf8(buf, -1, transcode_buf, BFBUFSIZ);
				transcode_buf[BFBUFSIZ-1] = '\0';
				utf8_buf = transcode_buf;
			} else {
				utf8_buf = buf;
			}
			q = newbuf;
			for (s = utf8_buf; *s; ++s) {
				if (strncmp(s, srch, srchlen) == 0) {
					memcpy(q, repl, repllen);
					q += repllen;
					s += srchlen - 1;
					++changes;
				}
				else
					*q++ = *s;
			}
			*q++ = '\0';
			if (strcmp(utf8_buf, newbuf)) {
				msg = calloc(1, strlen(utf8_buf) + strlen(newbuf) + 10);	
				format_message(msg, bfCat(M_STRSUB_BEFORE), utf8_buf);
				if ((s = strchr(msg, '\n')) != NULL)
					*s = 0;
				DataBlock(msg, NULL);
				format_message(msg, bfCat(M_STRSUB_AFTER), newbuf);
				if ((s = strchr(msg, '\n')) != NULL)
					*s = 0;
				DataBlock(msg, NULL);
				free(msg);
			}
			if (transcode) {
				convert_utf8_to_native(newbuf, -1, transcode_buf, BFBUFSIZ);
				utf8_buf = transcode_buf;
			} else {
				utf8_buf = newbuf;
			}
			if (fputs(utf8_buf, outf) == EOF) {
				fclose(inf);
				fclose(outf);
				format_message(buf, bfCat(E_STRSUB_REWRITE_ERROR), errno, path, tmpname);
				DataBlock(buf, NULL);
				free(tmpname);
				free(buf);
				free(newbuf);
				freeArgs(args);
				return 8;
			}
		}
		if (!feof(inf) && errno) {
			fclose(inf);
			fclose(outf);
			format_message(buf, bfCat(E_STRSUB_READ_TMP_ERROR), errno, tmpname);
			DataBlock(buf, NULL);
			free(tmpname);
			free(buf);
			free(newbuf);
			freeArgs(args);
			return 9;
		}
		fclose(inf);
		fclose(outf);
	
		format_message(buf, bfCat(M_STRSUB_SUMMARY), lines, changes);
		DataBlock(buf, NULL);
	
		/* clean up! */
		platform_unlink(tmpname);
		free(tmpname);
		free(buf);
		free(newbuf);
	}
	freeArgs(args);
	return 0;
}


