
#include "bf_agent.h"

#ifndef WIN32
extern int errno;
#endif

int FuncCCChanges(void);
int FuncFileChanges(char *, const char *);
int FuncDirChanges(char *, const char *);
int CCChanges(char *, const char *);
int CCChangesDir(char *, const char *);
void CCGenFileDiff(char *, char *, int);
void CCGenDirDiff(char *, char *, int);
char *acquire_branch_from_path(char *);

#ifdef WIN32
#define BF_NULL_DEVICE "nul:"
#else
#define BF_NULL_DEVICE "/dev/null"
#endif

static char *lnulldev = BF_NULL_DEVICE;

int cc_audit = 0;

int FuncFileChanges(char *path, const char *change_time) {
	char	*p, *buf, *findpath, *origpath, *tmpname, *segment, *find_opts_env;
	char    *bf_rptr=NULL;
	char    *bf_sptr=NULL;
	int     ccc_count=0;
	FILE	*f;

	if (NULL == (find_opts_env = platform_getenv("CC_FIND_OPTS")))
		find_opts_env = " ";

	buf = calloc(sizeof(*buf), BFBUFSIZ);
	findpath = calloc(sizeof(*findpath), BFBUFSIZ);
	tmpname = platform_tempnam();
	platform_unlink(tmpname);

	segment = bf_strtok(path, ",", &bf_rptr, &bf_sptr);
	while (segment) {
		format_message(findpath, "%s/%s/%s", this_view(), clearcase_view, segment);
		origpath = findpath;
		findpath = env_expand_var(findpath, 0, 0);
		free(origpath);
		CleanUnixPath(findpath);
		format_message(buf, "cleartool find %s -type f %s -cview -version \"created_since(%s)\" -print >> %s",
				findpath, find_opts_env, change_time, tmpname);
		BFDEBUG(buf);
		run_process(".", buf, 0, NULL, NULL, NULL, NULL, 0);
		segment = bf_strtok(NULL, ",", &bf_rptr, &bf_sptr);
	}

	/* parse what we got from the find */
	f = platform_fopen(tmpname, "r");
	if (f != NULL) {
		while (platform_fgets(buf, BFBUFSIZ, f) != NULL) {
			RemoveNewline(buf);
			CleanUnixPath(buf);
			if ((p = strrchr(buf, PathSeparator())) != NULL && *++p && atol(p) < 1)
				continue;
	                ccc_count++;
			CCChanges(buf, change_time);
		}
		fclose(f);
 		format_message(buf, bfCat(M_CC_CALLS), "FuncCCChanges", ccc_count, "CCChanges");
		BFDEBUG(buf);
	}

	free(buf);
	free(findpath);
	platform_unlink(tmpname);
	free(tmpname);
	platform_freeenv(find_opts_env);
	return 0;
}


int FuncDirChanges(char *path, const char *change_time) {
	char	*p, *buf, *findpath, *origpath, *tmpname, *segment, *find_opts_env;
	char    *bf_rptr=NULL;
	char    *bf_sptr=NULL;
	int     ccc_count=0;
	FILE	*f;

	if (NULL == (find_opts_env = platform_getenv("CC_FIND_OPTS")))
		find_opts_env = " ";

	buf = calloc(1, BFBUFSIZ);
	findpath = calloc(1, BFBUFSIZ);
	tmpname = platform_tempnam();
	platform_unlink(tmpname);
	segment = bf_strtok(path, ",", &bf_rptr, &bf_sptr);
	while (segment) {
		format_message(findpath, "%s/%s/%s", this_view(), clearcase_view, segment);
		origpath = findpath;
		findpath = env_expand_var(findpath, 0, 0);
		free(origpath);
		CleanUnixPath(findpath);
		format_message(buf, "cleartool find %s -type d %s -cview -version \"created_since(%s)\" -print >> %s",
				findpath, find_opts_env, change_time, tmpname);
		BFDEBUG(buf);
		run_process(".", buf, 0, NULL, NULL, NULL, NULL, 0);
		segment = bf_strtok(NULL, ",", &bf_rptr, &bf_sptr);
	}

	/* parse what we got from the find */
	f = platform_fopen(tmpname, "r");
	if (f != NULL) {
		while (platform_fgets(buf, BFBUFSIZ, f) != NULL) {
			RemoveNewline(buf);
			CleanUnixPath(buf);
			if ((p = strrchr(buf, PathSeparator())) != NULL && *++p && atol(p) < 1)
				continue;
	                ccc_count++;
			CCChangesDir(buf, change_time);
		}
		fclose(f);
 		format_message(buf, bfCat(M_CC_CALLS), "FuncDirChanges", ccc_count, "CCChanges");
		BFDEBUG(buf);
	}

	free(findpath);
	platform_unlink(tmpname);
	free(tmpname);
	free(buf);
	platform_freeenv(find_opts_env);
	return 0;
}

/* function to generate a diff report for a file */
void CCGenFileDiff(char *path, char *user, int blockdiff) {
	char 	*buf, *final_revision, *tfile, *diffs;
	struct 	stat sb;

	FILE	*f;

	final_revision = strdup(path);

	buf = calloc(1, BFBUFSIZ);
	tfile = platform_tempnam();
	diffs = NULL;
	CleanPath(final_revision);
	format_message(buf, "cleartool diff -pred -diff_format %s > %s 2> %s",
			final_revision, tfile, lnulldev);
	BFDEBUG(buf);
	run_process(".", buf, 0, NULL, NULL, NULL, NULL, 0);

	if (platform_stat(tfile, &sb) == 0 && sb.st_size > 0 && (f = platform_fopen(tfile, "r")) != NULL) {
		diffs = calloc(1, sb.st_size+8);
		fread(diffs, 1, sb.st_size, f);
		fclose(f);
		format_message(buf, "SRC_DIFF:predecessor:%s:%s", user, final_revision);
		CleanUnixPath(buf);
		DataBlock(diffs, buf);
	}
	platform_unlink(tfile);
	if (diffs)
		free(diffs);
	free(tfile);
	free(buf);
	free(final_revision);
}

/* function to generate a diff report for a dir */
void CCGenDirDiff(char *path, char *user, int blockdiff) {
	char 	*buf, *tfile, *diffs, *type_str = NULL, *change, *lpath, *tpath, **args;
	int 	add_count=0;
	int 	del_count=0;
	int 	del_flag=0;
	int 	add_flag=0;
	struct	stat	tsb;

	FILE	*f;

	change = calloc(1, BFBUFSIZ * 2);
	buf = calloc(1, BFBUFSIZ);
	lpath = strdup(path);
	tfile = platform_tempnam();
	diffs = NULL;
	CleanPath(path);
	format_message(buf, "cleartool diff -pred -diff_format %s > %s 2> %s",
			lpath, tfile, lnulldev);
	BFDEBUG(buf);
	run_process(".", buf, 0, NULL, NULL, NULL, NULL, 0);

	if (platform_stat(tfile, &tsb) == 0 && tsb.st_size > 0 && (f = platform_fopen(tfile, "r")) != NULL) {
		while (platform_fgets(buf, BFBUFSIZ, f) != NULL) {
			if (strstr(buf, "[ deleted ]")) {
				del_flag=1;
				continue;
                        }
			if (strstr(buf, "[ added ]")) {
				add_flag=1;
				continue;
                        }
			if (add_flag) {
				type_str = "SRC_ADD";
				add_count++;
			}
			if (del_flag) {
				type_str = "SRC_DELETE";
				del_count++;
			} 
			strcpy(change, buf);
			CleanUnixPath(change);
			args = getArgs(change);
			tpath = bf_strpart(lpath, "@@");
			format_message(change, "%s%c%s", tpath, PathSeparator(), args[1]);
			freeArgs(args);
			free(tpath);
			format_message(buf, "%s:predecessor:%s:%s:%s", type_str, user, lpath, change);
			DataBlock(NULL, buf);
			memset(change, 0, BFBUFSIZ * 2);
			add_flag=0;
			del_flag=0;
		}
		fclose(f);
 		format_message(buf, bfCat(M_CC_GENDIRDIFF), add_count, del_count);
		BFDEBUG(buf);
	}

	platform_unlink(tfile);
	free(tfile);
	free(change);
	free(lpath);
	free(buf);
}

char *acquire_branch_from_path(char *ppath) {
	char **largs, *lpath;
	int i=0;
	char *retstr=NULL;

	lpath = strdup(ppath);
	largs = getArgsC(lpath, PathSeparator());
	if (largs != NULL) {
		while (largs[i])
			i++;

		if (i>=2) {
			retstr = strdup(largs[i-2]);
		}
	}

	if (retstr == NULL)
		retstr = strdup(bfCat(E_CC_NO_PARSE_PATH));

	return retstr;
}

/* function to report a view diff */
int CCChanges(char *ppath, const char *change_time)
{
	char	*path, *tfile, *buf, *change, *basepath;
	char	*comment, *branch;
	int     ccc_count=0, len;
	struct	stat	tsb;
	FILE	*f;

	/* we have a VOB object that's changed since our reference time. */
	/* collect the change history information for this file */
	tfile = platform_tempnam();
	buf = calloc(1, BFBUFSIZ);
	change = calloc(1, BFBUFSIZ << 1);
	path = strdup(ppath);
	CleanPath(path);
	branch = acquire_branch_from_path(path);
	format_message(buf, "cleartool lshistory -nop -nco -fmt \"%%o:%%u:%%Nd:%%n\\n%%c\\n!!!\\n\" -since \"%s\" -branch %s %s > %s",
			change_time, branch, path, tfile);
	BFDEBUG(buf);
	run_process(".", buf, 0, NULL, NULL, NULL, NULL, 0);
	free(branch);
	if (platform_stat(tfile, &tsb) == 0 && tsb.st_size > 0 && (f = platform_fopen(tfile, "r")) != NULL) {
		while (platform_fgets(buf, BFBUFSIZ, f) != NULL) {
			if (strcmp(buf, "!!!\n") == 0) {
				if (strncmp(change, "checkin:", 8) == 0) {
					RemoveNewline(change);
					if ((comment = strchr(change, '\n'))) {
						*comment++ = 0;
					} else {
						comment = change;
					}
					CleanUnixPath(change);
					format_message(buf, "SRC_CHANGE:%s", change);
					DataBlock(comment, buf);
					if ((basepath = strchr(&change[8], ':')) != NULL) {
						*basepath++ = 0;
						if ((basepath = strchr(basepath, ':')) != NULL) {
							CCGenFileDiff(++basepath, &change[8], 0);
							ccc_count++;
						}
					}
				}
				memset(change, 0, (BFBUFSIZ << 1));
				continue;
			}
			len = strlen(change);
			if (len + strlen(buf) < (BFBUFSIZ << 1))
				strcpy(&change[len], buf);
		}
		fclose(f);
		format_message(buf, bfCat(M_CC_CALLS),
				"CCChanges", ccc_count, "CCGenFileDiff", ccc_count);
		BFDEBUG(buf);
	}
	free(buf);
	free(change);
	platform_unlink(tfile);
	free(tfile);
	free(path);
	return 0;
}

/* function to report a view dir diff */
int CCChangesDir(char *ppath, const char *change_time)
{
	char	*path, *tfile, *buf, *change, *basepath;
	char	*comment, *branch;
	int     ccc_count=0, len;
	struct	stat	tsb;
	FILE	*f;

	/* we have a VOB object that's changed since our reference time. */
	/* collect the change history information for this dir */
	tfile = platform_tempnam();
	buf = calloc(1, BFBUFSIZ);
	change = calloc(1, BFBUFSIZ << 1);
	path = strdup(ppath);
	CleanPath(path);
	branch = acquire_branch_from_path(path);
	format_message(buf, "cleartool lshistory -directory -nop -nco -fmt \"%%o:%%u:%%Nd:%%n\\n%%c\\n!!!\\n\" -since \"%s\" -branch %s %s > %s",
			change_time, branch, path, tfile);
	BFDEBUG(buf);
	run_process(".", buf, 0, NULL, NULL, NULL, NULL, 0);
	free(branch);
	if (platform_stat(tfile, &tsb) == 0 && tsb.st_size > 0 && (f = platform_fopen(tfile, "r")) != NULL) {
		while (platform_fgets(buf, BFBUFSIZ, f) != NULL) {
			if (strcmp(buf, "!!!\n") == 0) {
				if (strncmp(change, "checkin:", 8) == 0) {
					RemoveNewline(change);
					if ((comment = strchr(change, '\n')))
						*comment++ = 0;
					else
						comment = change;
					CleanUnixPath(change);
					format_message(buf, "SRC_CHANGE:%s", change);
					DataBlock(comment, buf);
					if ((basepath = strchr(&change[8], ':')) != NULL) {
						*basepath++ = 0;
						if ((basepath = strchr(basepath, ':')) != NULL) {
							CCGenDirDiff(++basepath, &change[8], 0);
	                                                ccc_count++;
	                                        }
					}
				}
				memset(change, 0, BFBUFSIZ << 1);
				continue;
			}
			len = strlen(change);
			if (len + strlen(buf) < (BFBUFSIZ << 1))
				strcpy(&change[len], buf);
		}
		fclose(f);
 		format_message(buf, bfCat(M_CC_CALLS), "CCChangesDir", ccc_count, "CCGenDirDiff");
		BFDEBUG(buf);
	}
	free(buf);
	free(change);
	platform_unlink(tfile);
	free(tfile);
	free(path);
	return 0;
}

/*
 * read a local file/tree and send it to stdout.
 *
 */
int FuncCCChanges(void)
{
	const char *path, *stamp;
	char	*spath;
	int 	mapfail;

	/*  locate the path element */
	if (NULL == (path = request_get(TAG_PATH))) {
		ResultBlock(E_READ_NOPATH, -1, bfCat(E_READ_NOPATH));
		return 0;
	}
        spath = strdup(path);

	if (NULL == (stamp = request_get(TAG_STAMP))) {
		ResultBlock(E_READ_NOPATH, -1, bfCat(E_READ_NOPATH));
		return 0;
	}

	if (!clearcase_view) {
		ResultBlock(E_CCASE_NOVIEW, -1, bfCat(E_CCASE_NOVIEW));
		return 0;
	}

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		ResultBlock(E_EXEC_MAPFAIL, -1, bfCat(E_EXEC_MAPFAIL));
		return 0;
	}

	ClobberPasswords();

	/* Format is now done on the console side; same format */
	/* stamp = formatUTCDateStamp("%d-%b-%Y.%H:%M:%SUTC", atol(stamp)); */
	/* if (1 || !FuncFileChanges(clearcase_view, spath, stamp)) */
	if (!FuncFileChanges(spath, stamp))
		FuncDirChanges(spath, stamp);

	ResultBlock(0, -1, "");
	platform_unmap_drives();
	free(spath);
	return 0;
}

/*
 *
 * function to start clearcase views when _CLEARCASE_VIEWS is set
 * returns an alloc'd buffer that contains a valid pathname for 
 * the first view started.
 *
 */
char	*first_cc_view = NULL;
char *CCStartViews(int timeout) {
	char	*viewlist, *view, *sptr, *rptr;
	char	*cmd;
	int 	timeout_result, rc;


	if (NULL == (viewlist = platform_getenv("_CLEARCASE_VIEWS")))
		return NULL;

	if (first_cc_view)
		free(first_cc_view);
	first_cc_view = NULL;
	sptr = rptr = NULL;

	/* we'll build a list of things to do in the cmd buffer and feed it to run_process() */
	cmd = calloc(1, BFBUFSIZ);
	rc = BFBUFSIZ;

	view = bf_strtok(viewlist, ";:,", &rptr, &sptr);
	while (view) {
		/* if the buffer isn't big enough (that's a LOT of views!) grow it */
		if ((size_t) rc < strlen(cmd) + (strlen(view) * 2) + 55) {
			rc += BFBUFSIZ;
			cmd = realloc(cmd, rc);
		}
		format_message(&cmd[strlen(cmd)],
 				"echo EXEC: cleartool startview %s\ncleartool startview %s\n",
				view, view);
		if (!first_cc_view) {
			char *msg = calloc(1, BFBUFSIZ);
			first_cc_view = calloc(1, BFBUFSIZ);
			format_message(first_cc_view, "%s/%s", this_view(), view);
			CleanPath(first_cc_view);
			format_message(msg, bfCat(E_CCASE_STARTVIEW_PATH), first_cc_view);
			DataBlock(msg, NULL);
			free(msg);
		}
		view = bf_strtok(NULL, ";:,", &rptr, &sptr);
	}
	bf_strtok(NULL, NULL, NULL, &sptr);

	/* the command input script has been generated, go run it! */
	if (*cmd != '\0')
		rc = run_process(".", cmd, timeout, NULL, NULL, NULL, &timeout_result, 0);

	free(cmd);

	/* free the original env pointer if necessary */
	platform_freeenv(viewlist);
	return first_cc_view;
}

/*
 *
 * function to mount clearcase vobs when _CLEARCASE_VOBS is set
 *
 */
int CCMountVobs(int timeout) {
	char	*voblist, *vob, *sptr, *rptr, *cmd, *sbuf;
	int 	rc, timeout_result;
	struct stat	sb;

	if (NULL == (voblist = platform_getenv("_CLEARCASE_VOBS")))
		return -1;

	sptr = rptr = NULL;
	vob = bf_strtok(voblist, ";:,", &rptr, &sptr);
	cmd = calloc(1, BFBUFSIZ);
	sbuf = calloc(1, BFBUFSIZ);
	rc = BFBUFSIZ;

	while (vob) {
		/* if the buffer isn't big enough (that's a LOT of vobs!) grow it */
		if ((size_t)rc < strlen(cmd) + (strlen(vob) * 2) + 55) {
			rc += BFBUFSIZ;
			cmd = realloc(cmd, rc);
		}

		if (first_cc_view) {
			format_message(sbuf, "%s/%s/%s", this_view(), first_cc_view, vob);
			CleanPath(sbuf);
		}
		if (!first_cc_view || stat(sbuf, &sb) != 0) {
			/* mount the vob identified by 'vob' */
			format_message(&cmd[strlen(cmd)],
 					"echo EXEC: cleartool mount %s\ncleartool mount %s\n",
					vob, vob);
		}
		vob = bf_strtok(NULL, ";:,", &rptr, &sptr);
	}
	bf_strtok(NULL, NULL, NULL, &sptr);
	free(sbuf);

	/* the command input script has been generated, go run it! */
	if (*cmd != '\0')
		rc = run_process(".", cmd, timeout, NULL, NULL, NULL, &timeout_result, 0);

	/* free the original env pointer if necessary */
	free(cmd);
	platform_freeenv(voblist);
	return rc;
}
