/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

extern bf_internal_t internals[];
extern char *sourcelist;

int FuncExec(void)
{
	char	buf[BFBUFSIZ], msg[BFBUFSIZ], *p, *rptr, *sptr;
	char	*exec, *odir, *dir, *tag, *root, *timeout, *exp_exec = NULL, *indata;
	char	*ccroot = NULL;
	int 	result=0, i, timeout_occurred;
	int 	mapfail, no_preparse;
	char	*userdata = NULL;
	size_t	userdata_size;

	/* Handle the request parameters */
	if (NULL == (exec = request_get(TAG_EXEC))) {
		ResultBlock(E_EXEC_NOEXEC, -1, bfCat(E_EXEC_NOEXEC));
		return 0;
	}
	if (NULL == (odir = request_get(TAG_DIR))) {
		request_put(TAG_DIR,"");
		odir = request_get(TAG_DIR);
	}
	if (NULL == (tag = request_get(TAG_TAG))) {
		request_put(TAG_TAG,"");
		tag = request_get(TAG_TAG);
	}
	if (NULL == (root = request_get(TAG_ROOT))) {
		request_put(TAG_ROOT,"");
		root = request_get(TAG_ROOT);
	}
	if (NULL == (timeout = request_get(TAG_TIMEOUT)))
		timeout = "0";
	verbose = !request_has(TAG_SILENT);
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);
	indata = request_get(TAG_STDIN);

	/* From this point forward, we are allocating dynamic data structures
	 * that need to be cleaned back up when we're done.
	 */

	/*  build the working directory path */
	CleanPath(root);
	CleanPath(tag);
	CleanPath(odir);
	root = env_expand_var(root, 0, 0);
	dir = env_expand_var(odir, 0, 0);
	CleanPath(dir);
	if (strcmp(dir, odir)) {
		format_message(msg, bfCat(E_EXEC_XLATE), root, tag, odir, root, tag, dir);
		CleanPath(msg);
		DataBlock(msg, NULL);
	}

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		strcpy(buf, bfCat(result = E_EXEC_MAPFAIL));
		goto done;
	}

	ClobberPasswords();

	if (NULL != (p = platform_getenv("_TIMEOUT"))) {
		timeout = p;
		format_message(buf, bfCat(E_EXEC_TIMEOUT_OVERRIDE), atoi(timeout));
		DataBlock(buf, NULL);
	}



	/* we don't prefix the clearcase view root here because this path */
	/* will be used under a setview shell which will make the path valid */
	/* we just need to postpone validating the directory until later */
	/* we'll go ahead and build the ccroot for the internals, however */
	ccroot = NULL;
	if (clearcase_view) {
		ccroot = malloc(BFBUFSIZ);
		format_message(ccroot, "%s/%s/%s/%s/%s",
				this_view(), clearcase_view, root, tag, dir);
		CleanPath(ccroot);
	}

	format_message(buf, "%s/%s/%s", root, tag, dir);
	CleanPath(buf);

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/*  make sure we can cd to it. */
	errno = 0;
	if (ccroot) {
		if (platform_chdir(ccroot) != 0) {
			format_message(msg, bfCat(result = E_EXEC_CDFAIL), ccroot, errno);
			goto done;
		}
	} else {
		if (platform_chdir(buf) != 0) {
			format_message(msg, bfCat(result = E_EXEC_CDFAIL), buf, errno);
			goto done;
		}
	}

	OpenLog();
	DumpEnv();

	no_preparse = BFCONF_EXISTS(no_preparse_command);
	if (!no_preparse) {
		p = platform_getenv("_NO_PREPARSE_COMMAND");
		if (p) {
			platform_freeenv(p);
			no_preparse = 1;
		}
	}

	timeout_occurred = 0;

	/* see if exec is an internal command */
	/* we pass the unexpanded version of the command string */
	for (i = 0 ; internals[i].cmd && strncmp(internals[i].cmd, exec, strlen(internals[i].cmd)) ; i++);
	if (internals[i].cmd) {
		format_message(msg, bfCat(E_EXEC_RUN), buf, platform_hostname(), exec);
		DataBlock(msg, NULL);

		if (!no_preparse) {
			DataBlock(bfCat(E_EXEC_PREPARSE), NULL);
			exp_exec = env_expand_hidden(exec, 0, 0);
		}
		else
			exp_exec = strdup(exec);

		p = &exp_exec[strlen(internals[i].cmd)];
		while (*p && isspace((int)(*p)))
			++p;

		result = (internals[i].func)(p, atoi(timeout), &timeout_occurred);
	}
	else if (bf_dynamic_function(exec, atoi(timeout), &result, &timeout_occurred)) {
		exp_exec = strdup(exec);
	}
	else {
		format_message(msg, bfCat(E_EXEC_RUN), buf, platform_hostname(), exec);
	/* unix systems need to allocate a pty, so we delay the output of the start message */
	/* until inside run_process() on unix */
#ifdef WIN32
		DataBlock(msg, NULL);
#endif
		if (!no_preparse) {
			DataBlock(bfCat(E_EXEC_PREPARSE), NULL);
			exp_exec = env_expand_hidden(exec, 0, 0);
			if (strcmp(exp_exec, exec) != 0) {
				format_message(msg, bfCat(E_EXEC_RUN), buf, platform_hostname(), exp_exec);
#ifdef WIN32
				DataBlock(msg, NULL);
#endif
			}
		}
		else
			exp_exec = strdup(exec);

		result = run_process(
				buf,
				exp_exec,
				atoi(timeout),
				indata,
				sourcelist,
				msg,
				&timeout_occurred,
				(trapCommand && strlen(trapCommand)) ? 0 : 1);
	}

	/* allow other codes to be treated as a pass */
	if (result && NULL != (p = platform_getenv("_EXITCODE_MAP"))) {
		strcpy(msg, p);
		rptr = sptr = NULL;
		p = bf_strtok(msg, ",:; ", &rptr, &sptr);
		while (p && result) {
			i = atoi(p);
			if (result == i) {
				format_message(msg, bfCat(E_EXEC_EXITMAP), result);
				DataBlock(msg, NULL);
				result = 0;
			}
			p = bf_strtok(NULL, ",:; ", &rptr, &sptr);
		}
		bf_strtok(NULL, NULL, NULL, &sptr);
	}

	if (result && trapCommand && strlen(trapCommand)) {
		format_message(msg, bfCat(E_EXEC_TRAP), result, trapCommand);
		DataBlock(msg, NULL);
		trapmode = 1;
		run_process(buf, trapCommand, 60, NULL, NULL, NULL, &timeout_occurred, 1);
		trapmode = 0;
	}

/*	p = bf_expose_special_chars(exp_exec); */
	if (timeout_occurred) {
		format_message(msg, bfCat(E_EXEC_TIMEOUT), timeout_occurred);
	} else {
		format_message(msg, bfCat(E_EXEC_OK), buf, platform_hostname(), exp_exec);
	}
/*	free(p); */

done:
	platform_unmap_drives();
	FreeUserData();
	ResultBlock(result, -1, msg);
	free(exp_exec);
	free(ccroot);
	free(dir);
	free(root);
	return 0;
}

