/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/*
 * This needs to be smaller than BFBUFSIZ, as otherwise we
 * could overflow the msg buffer.
 */
#define EXPAND_BUFSIZ (BFBUFSIZ >> 1)

extern char *sourcelist;

int FuncExpand(void)
{
	char	buf[BFBUFSIZ], msg[BFBUFSIZ], *p, *tmpfilename = NULL, *shell_format, *shell;
	char	*exp_exec = NULL, *exec, *odir, *dir = NULL, *root, *timeout;
	char	*ccroot = NULL, *userdata;
	int 	result, count, timeout_occurred;
	int 	fd, mapfail, no_preparse;
	struct	stat	sb;
	size_t	userdata_size;
#ifndef WIN32
	struct	passwd	*pwd;
#endif

	/* Handle the request parameters */
	if (NULL == (exec = request_get(TAG_EXEC))) {
		ResultBlock(E_EXEC_NOEXEC, -1, bfCat(E_EXEC_NOEXEC));
		return 0;
	}
	if (NULL == (odir = request_get(TAG_DIR))) {
		request_put(TAG_DIR, "");
		odir = request_get(TAG_DIR);
	}
	if (NULL == (root = request_get(TAG_ROOT))) {
		request_put(TAG_ROOT, "");
		root = request_get(TAG_ROOT);
	}
	if (NULL == (timeout = request_get(TAG_TIMEOUT)))
		timeout = "0";
	verbose = !request_has(TAG_SILENT);
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		strcpy(buf, bfCat(result = E_EXEC_MAPFAIL));
		goto done;
	}

	ClobberPasswords();

	if (NULL != (p = platform_getenv("_TIMEOUT"))) {
		timeout = p;
		format_message(buf, bfCat(E_EXEC_TIMEOUT_OVERRIDE), atoi(timeout));
		DataBlock(buf, NULL);
	}

	CleanPath(root);
	CleanPath(odir);

	/*  build the working directory path */
	root = env_expand_var(root, 0, 0);
	dir = env_expand_var(odir, 0, 0);
	CleanPath(dir);
	if (strcmp(dir, odir)) {
		format_message(msg, bfCat(E_EXEC_XLATE), root, "", odir, root, "", dir);
		CleanPath(msg);
		DataBlock(msg, NULL);
	}

	/* we don't prefix the clearcase view root here because this path */
	/* will be used under a setview shell which will make the path valid */
	/* we just need to postpone validating the directory until later */
	/* we'll go ahead and build the ccroot for the internals, however */
	ccroot = NULL;
	if (clearcase_view) {
		ccroot = malloc(BFBUFSIZ);
		format_message(ccroot, "%s/%s/%s/%s", this_view(), clearcase_view, root, dir);
		CleanPath(ccroot);
	}

	format_message(buf, "%s/%s", root, dir);
	CleanPath(buf);

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/*  make sure we can cd to it. */
	errno = 0;
	if (ccroot) {
		if (platform_chdir(ccroot) != 0) {
			format_message(msg, bfCat(result = E_EXEC_CDFAIL), ccroot, errno);
			goto done;
		}
	}
	else {
		if (platform_chdir(buf) != 0) {
			format_message(msg, bfCat(result = E_EXEC_CDFAIL), buf, errno);
			goto done;
		}
	}

	DumpEnv();

	/* we need to determine the user's shell. */
	shell_format = BOURNE_REDIRECT;
	shell = NULL;
#ifndef WIN32
	if ((shell = bfconf_get("shell", NULL)) != NULL) {
		CleanUnixPath(shell);
		if ((p = strrchr(shell, '/')) != NULL)
			strcpy(shell, ++p);
	}
	else {
		if ((pwd = getpwuid(getuid())) != NULL) {
			shell = strdup(pwd->pw_shell);
			if ((p = strrchr(shell, '/')) != NULL)
				strcpy(shell, ++p);
		}
		else
			shell = strdup("sh"); /* I give up */
		endpwent();
	}

	/* shell now points to the base shell name. figure out what it is and */
	/* then free it */
	if (strcmp(shell, "csh") == 0 || strcmp(shell, "tcsh") == 0)
		shell_format = CSH_REDIRECT;
	if (shell)
		free(shell);
#endif

	no_preparse = BFCONF_EXISTS(no_preparse_command);
	if (!no_preparse) {
		p = platform_getenv("_NO_PREPARSE_COMMAND");
		if (p) {
			platform_freeenv(p);
			no_preparse = 1;
		}
	}

	if (no_preparse)
		exp_exec = strdup(exec);
	else
		exp_exec = env_expand_var(exec, 0, 0);

	/* add redirect for exec to a tmp file */
	tmpfilename = platform_tempnam();
	platform_unlink(tmpfilename);
	tmpfilename = realloc(tmpfilename, strlen(tmpfilename) + 5);
	format_message(tmpfilename + strlen(tmpfilename), ".out");
	p = calloc(1, strlen(exp_exec) + strlen(tmpfilename) + 16);
#ifdef WIN32
	if (cygwin)
		CleanUnixPath(tmpfilename);
#endif
	format_message(p, shell_format, exp_exec, tmpfilename);
	free(exp_exec);
	format_message(msg, bfCat(E_EXEC_RUN), buf, platform_hostname(), exp_exec = p);

	/* unix systems need to allocate a pty, so we delay the output of the start message */
	/* until inside run_process() */
#ifdef WIN32
	DataBlock(msg, NULL);
#endif

	timeout_occurred = 0;
	result = run_process(buf, exp_exec, atoi(timeout), NULL, sourcelist, msg, &timeout_occurred, 1);

	/* load in tmp file */
	if (0 != platform_stat(tmpfilename, &sb) || sb.st_size > EXPAND_BUFSIZ)
		sb.st_size = EXPAND_BUFSIZ;  /* don't read any more than this */

	fd = platform_open2(tmpfilename, O_RDONLY);
	if (fd > -1) {
		count = platform_read(fd, buf, sb.st_size);
		platform_close(fd);

		if (count < 0)
			count = 0;
		else if (count >= EXPAND_BUFSIZ)
			count = EXPAND_BUFSIZ - 1;
		buf[count++] = '\0';
	} else {
		buf[0] = '\0';
		count = 1;
	}

	if (!platform_utf8_is_native()) {
		char *x = malloc(count);
		memcpy(x, buf, count);
		count = convert_native_to_utf8(x, count, buf, EXPAND_BUFSIZ);
		buf[EXPAND_BUFSIZ - 1] = '\0';
	}

	/* if buf ends in a newline, strip it. ResultBlock will add one */
	RemoveNewline(buf);

	platform_unlink(tmpfilename);

	if (timeout_occurred)
		format_message(msg, bfCat(E_EXEC_TIMEOUT), timeout_occurred);
	else
		format_message(msg, bfCat(E_EXEC_OK), buf, platform_hostname(), exp_exec);

done:
	platform_unmap_drives();
	FreeUserData();
	ResultBlock(result, strlen(buf), buf);
	free(exp_exec);
	free(root);
	free(dir);
	free(tmpfilename);
	free(ccroot);
	return 0;
}

