

#include "bf_agent.h"

/* Define these so we won't have to pull in platform.c */
char *bfconf_get(const char *key, const char *def) { return NULL; }
const char *platform_setlocale(const char *locale) { return setlocale(LC_ALL, locale); }
#ifdef WIN32
void platform_set_language(USHORT primary, USHORT sub) { /* ignore */ }
#endif

/* We only want format_message ... */
#define LANG_TEST
#include "string.c"

/* Some macros to make specifying the tests a bit easier ... */
const char *format_result = "%3d: %-32s => {{{%s}}}\n";
const char *format_unused = "%3d: Unused\n";
#define V(code,name) \
	if (verbose) printf(format_result, code, name, buf)
#define X0(code) \
	format_message(buf, bfCat(code)); \
	V(code, #code)
#define X1(code, arg1) \
	format_message(buf, bfCat(code), arg1); \
	V(code, #code)
#define X2(code, arg1, arg2) \
	format_message(buf, bfCat(code), arg1, arg2); \
	V(code, #code)
#define X3(code, arg1, arg2, arg3) \
	format_message(buf, bfCat(code), arg1, arg2, arg3); \
	V(code, #code)
#define X4(code, arg1, arg2, arg3, arg4) \
	format_message(buf, bfCat(code), arg1, arg2, arg3, arg4); \
	V(code, #code)
#define X5(code, arg1, arg2, arg3, arg4, arg5) \
	format_message(buf, bfCat(code), arg1, arg2, arg3, arg4, arg5); \
	V(code, #code)
#define X6(code, arg1, arg2, arg3, arg4, arg5, arg6) \
	format_message(buf, bfCat(code), arg1, arg2, arg3, arg4, arg5, arg6); \
	V(code, #code)
#define X7(code, arg1, arg2, arg3, arg4, arg5, arg6, arg7) \
	format_message(buf, bfCat(code), arg1, arg2, arg3, arg4, arg5, arg6, arg7); \
	V(code, #code)
#define UNUSED(code) \
	if (verbose) printf(format_unused, code)

static const char *langs[] = {
	"de_DE",
	"en_US",
	"es_ES",
	"fr_FR",
	"it_IT",
	"ja_JP",
	"ko_KR",
	"pt_BR",
	"zh_CN",
	"zh_TW",
	NULL
};

#ifdef WIN32
static const char *win_locale[] = {
	"German_Germany.1252",
	"English_United States.1252",
	"Spanish_Spain.1252",
	"French_France.1252",
	"Italian_Italy.1252",
	"Japanese_Japan.932",
	"Korean_Korea.949",
	"Portuguese_Brazil.1252",
	"Chinese_China.936",
	"Chinese_Taiwan.950",
	NULL
};
#endif

static const char *conf = "#CONF_FILE#";
static const char *sidtype = "#SID_TYPE#";
static const char *domain = "#DOMAIN#";
static const char *user = "#USER#";
static const char *host = "#HOST#";
static const char *priv = "#PRIVILEGE#";
static const char *cmd = "#COMMAND_LINE#";
static const char *fulldir = "#FULL_DIR#";
static const char *root = "#ROOT#";
static const char *tag = "#TAG#";
static const char *dir = "#DIR#";
static const char *dirvar = "#DIR_WITH_VARS#";
static const char *tmpf = "#TMP_FILE#";
static const char *srcf = "#SRC_FILE#";
static const char *logf = "#LOG_FILE#";
static const char *pty = "#PTY#";
static const char *locale = "#LOCALE#";
static const char *view = "#CC_VIEW#";
static const char *vob = "#CC_VOB#";
static const char *dlfn = "#DYN_FN#";
static const char *dllib = "#DYN_LIB#";
static const char *monf = "#MONITOR_FILE#";
static const char *mapdrv = "#MAP_DRIVE#";
static const char *mapdir = "#MAP_DIRECTORY#";
static const char *mapvia = "#MAP_PROVIDER#";
static const char *timestr = "#TIME_DISPLAY#";
static const char *editsrch = "/#SEARCH#";
static const char *editrepl = "#REPLACE#";
static const char *reqdisk = "#REQUIRE_DISK#";
static const char *reqmem = "#REQUIRE_MEM#";
static const char *shell = "#SHELL#";
static const char *hidden = "#HIDDEN#";
static const char *env = "#ENV_VAR=ENV_VALUE#";
static const char *envvar = "#ENV_VAR#";
static const char *envvalue = "#ENV_VALUE#";
static const char *envprev = "#ENV_PREV#";
static const char *envdotcmd = "#ENV_DOT_CMD#";
static const char *envdotcmdarg = "#ENV_DOT_CMD_ARG#";
static const char *sslcert = "#SSL_CERT#";
static const char *sslcipher = "#SSL_CYPHER#";
static const char *sslissuer = "#SSL_ISSUER#";
static const char *project = "#PROJECT#";
static const char *step = "#STEP#";
static const char *cccaller = "#CC_CALLER#";
static const char *cccallee = "#CC_CALLEE#";
static const void *ptr = (void *)((unsigned int)0xDEADF00D);


static void test_messages(int verbose) {
	char buf[BFBUFSIZ];

	X0(CODE_NOCOMMAND);
	X0(CODE_REQUESTDISCARDED);
	X0(CODE_BADCOMMAND);
	X0(CODE_UNKNOWN);
	X0(E_TOOMANYTAGS);
	X0(E_NOLICENSE);
	X0(E_BADLICENSE);
	X3(E_EXEC_OK, user, host, cmd);
	X0(E_EXEC_NOEXEC);
	X0(E_EXEC_NODIR);
	X0(E_EXEC_NOTAG);
	X0(E_EXEC_NOROOT);
	X2(E_EXEC_CDFAIL, fulldir, 2);
	X1(E_EXEC_PIPEFAIL, cmd);
	X0(E_EXEC_NOTIMEOUT);
	X0(E_EXEC_TIMEOUT_POP);
	X1(E_EXEC_TIMEOUT, 2);
	X3(E_EXEC_RUN, user, host, cmd);
	X1(E_EXEC_EXECFAIL, 2);
	X6(E_EXEC_XLATE, root, tag, dirvar, root, tag, dir);
	X2(E_SETUSER_FAIL, user, 1326);
	X1(M_SETUSER_OK, user);
	X2(E_EXEC_SCRIPTFAIL, 5, tmpf);
	X2(E_EXEC_COPYFAIL, srcf, 2);
	X1(E_EXEC_COPY, srcf);
	X2(E_EXEC_NOPTY, pty, 2);
	X1(E_EXEC_PTY, pty);
	X2(E_EXEC_NOLOCALE, locale, 2);
	X1(E_EXEC_LOCALE, locale);
	X1(E_EXEC_TRAP, 0);
	X1(E_MKDIR_OK, fulldir);
	X1(E_MKDIR_EXISTS, fulldir);
	X0(E_MKDIR_NODIR);
	X1(E_MKDIR_FAIL, fulldir);
	X0(E_MKDIR_NOROOT);
	X4(E_MKDIR_XLATE, root, dirvar, root, dir);
	UNUSED(36);
	UNUSED(37);
	X2(E_MKDIR_CDROOT, fulldir, 2);
	X1(E_RMDIR_OK, fulldir);
	X1(E_RMDIR_NOTDIR, fulldir);
	X0(E_RMDIR_NODIR);
	X1(E_RMDIR_FAIL, fulldir);
	X1(E_RMDIR_FAIL2, fulldir);
	X1(E_RMDIR_FAIL3, fulldir);
	X0(E_RMDIR_NOROOT);
	X4(E_RMDIR_XLATE, root, dirvar, root, dir);
	X0(E_RMDIR_DOTDOT);
	UNUSED(48);
	UNUSED(49);
	X1(E_RMDIR_NOTTHERE, fulldir);
	X1(E_READ_OK, tmpf);
	X1(E_READ_NOEXIST, tmpf);
	X1(E_READ_NOTADIR, tmpf);
	X1(E_READ_NOOPEN, tmpf);
	X1(E_READ_NOOPEN2, tmpf);
	X2(E_READ_FAIL, 2, tmpf);
	UNUSED(57);
	UNUSED(58);
	X0(E_READ_NOPATH);
	X1(E_READ_NOTFILE, tmpf);
	X2(E_WRITE_OK, 4096, tmpf);
	X0(E_WRITE_NOPATH);
	X1(E_WRITE_EXISTS, tmpf);
	X1(E_WRITE_NOOPEN, tmpf);
	X1(E_WRITE_NODATA, tmpf);
	X1(E_WRITE_FAIL, tmpf);
	UNUSED(67);
	UNUSED(68);
	X0(E_WRITE_NOOFFSET);
	X1(E_WRITE_WRITEFAIL, 28);
	X1(E_SIZE_OK, fulldir);
	X0(E_SIZE_NODIR);
	X0(E_SIZE_NOOPEN);
	X1(E_SIZE_NOEXIST, fulldir);
	X1(E_SIZE_FAIL, fulldir);
	X4(E_SIZE_XLATE, root, dirvar, root, dir);
	UNUSED(77);
	UNUSED(78);
	X1(E_CCASE_OK, view);
	X0(E_CCASE_NOPATH);
	UNUSED(81);
	UNUSED(82);
	X0(E_CCASE_NOVIEW);
	X0(E_CCASE_STAMP);
	X1(E_EXEC_CCVIEW, view);
	X1(E_EXEC_CCVIEW_WIN, view);
	X2(E_EXEC_CCSVTIMEOUT, view, 2);
	X2(E_EXEC_CCSVERROR, 1326, view);
	X1(E_RMDIR_NOTTHERE_EACCESS, fulldir);
	X1(E_RMDIR_NOTTHERE_ENOENT, fulldir);
	X1(E_RMDIR_NOTTHERE_ENOTDIR, fulldir);
	X1(E_SU_ILLEGAL, user);
	X1(E_SU_UNKNOWNUSER, user);
	X1(E_SU_ACCOUNTINVALID, user);
	X1(E_SU_SETUSERFAILED, user);
	X1(E_SU_USERNOTFOUND, user);
	X1(E_SU_PAMSESSIONERROR, user);
	X1(E_SU_NOCRED, user);
	X1(E_SU_NOSESSION, user);
	X1(E_SU_SYSTEMERROR, user);
	X0(P_LOOPBACK_OK);
	X2(E_EXEC_COPYFAIL_MSG, srcf, 2);
	X1(E_EXEC_COPY_MSG, srcf);
	X1(E_EXEC_CDFAIL_MSG, fulldir);
	X1(E_EXEC_CCVIEW_MSG_WIN, view);
	X1(E_EXEC_EXECFAIL_MSG, 1326);
	X1(E_EXEC_TIMEOUT_MSG, 300);
	X2(E_EXEC_DL_OK, dlfn, dllib);
	X3(E_EXEC_DL_NOPIPE, 5, dlfn, dllib);
	X3(E_EXEC_DL_NOFORK, 5, dlfn, dllib);
	X1(E_EXEC_TIMEOUT_OVERRIDE, 900);
	X0(E_ENV_UNDEF);
	X0(E_ENV_RECURSION_SUPPRESSED);
	X0(E_EXEC_MAPFAIL);
	X0(E_MONITOR_USAGE);
	X1(E_MONITOR_WAITINGFOR, monf);
	X1(E_MONITOR_TIMEOUT, monf);
	X1(E_MONITOR_STATFAIL, monf);
	X2(E_MONITOR_SIZEOK, monf, 1024);
	X3(E_MONITOR_STOPPED, monf, 4096, 30);
	X1(E_CTOOL_EXECFAIL, view);
	X1(E_FORK_STREAMFAIL, 2);
	X2(E_WIN_MAP_ASSIGNED, mapdrv, mapdir);
	X3(E_WIN_MAP_OK, mapdrv, mapdir, mapvia);
	X3(E_WIN_MAP_ERROR, 1326, mapdir, mapdrv);
	X2(E_WIN_MAP_NOLETTER, mapdir, mapdrv);
	X2(E_WIN_MAP_ATTEMPT, mapdir, mapdrv);
	X4(E_WIN_MAP_UNABLE, mapdir, mapdrv, 2, 1326);
	X3(E_WIN_MAP_OK_DEFAULT, mapdir, mapdrv, 24);
	X3(E_WIN_MAP_OK_CREDS, mapdir, mapdrv, user);
	X2(E_WIN_MAP_OK_SYSTEM, mapdir, mapdrv);
	X1(E_WIN_MAP_UNMAPPED, mapdrv);
	X2(E_WIN_MAP_UNMAPFAIL, 108, mapdrv);
	X3(E_WIN_AUTH_RUNNINGAS, user, domain, sidtype);
	X1(E_WIN_AUTH_PRIV, priv);
	X0(E_WIN_AUTH_NOPROFILE);
	X0(E_WIN_EXEC_JOBFAIL);
	X0(E_WIN_EXEC_PIPEFAIL);
	X0(E_WIN_EXEC_STREAMFAIL);
	X0(E_WIN_EXEC_IPIPEFAIL);
	X0(E_WIN_EXEC_NOCWD);
	X0(E_WIN_EXEC_JOBASSOCFAIL);
	X0(E_WIN_TMPNAMFAIL);
	X1(E_EXEC_EXITMAP, 42);
	X2(E_CONFOPENFAIL, conf, 2);
	X2(E_WRITE_SETMODE, tmpf, 0644);
	X2(E_WRITE_NOSETMODE, tmpf, 0644);
	X2(E_WRITE_SETTIME, tmpf, timestr);
	X2(E_WRITE_NOSETTIME, tmpf, timestr);
	X1(E_CCASE_STARTVIEW, view);
	X1(E_CCASE_MOUNTVOB, vob);
	X1(E_CCASE_STARTVIEW_OK, view);
	X1(E_CCASE_STARTVIEW_PATH, view);
	X1(E_CCASE_MOUNTVOB_OK, vob);
	X0(E_EXEC_PREPARSE);
	X1(E_EXEC_WARNOUT, 180);
	X0(WIN_EVENT_CTLC);
	X0(WIN_EVENT_CLOSE);
	X0(WIN_EVENT_BREAK);
	X0(WIN_EVENT_SHUTDOWN);
	X0(WIN_EVENT_LOGOFF);
	X0(E_EDIT_WRONG_ARGS);
	X0(E_EDIT_BAD_EXPRESSION);
	X5(M_EDIT_APPLYING, editsrch, *editsrch, editrepl, *editsrch, tmpf);
	X3(E_WRITE_FAILORIG, srcf, 5, tmpf);
	X2(E_READ_FAILORIG, 2, srcf);
	X2(M_EDIT_RESULTS, 100, 3);
	X0(E_EDIT_TERM_SEARCH);
	X0(E_EDIT_TERM_REPLACE);
	X2(E_RMDIR_RMERR, 5, fulldir);
	X0(M_SYSCONF_COMPLETE);
	X0(M_REQUIRE_NEGATING);
	X0(E_REQUIRE_DISK_UNSUPP);
	X1(E_REQUIRE_DISK_MALFORMED, reqdisk);
	X1(E_REQUIRE_DISK_BAD_PATH, reqdisk);
	X3(E_REQUIRE_DISK_INSUFF, reqdisk, 2048L, 4096L);
	X3(M_REQUIRE_DISK_OK, reqdisk, 8192L, 4096L);
	X1(E_REQUIRE_HUH, reqdisk);
	X0(M_SSL_PLAINTEXT);
	X1(M_TMP_FILE_IS, tmpf);
	X0(E_REQUIRE_MEM_UNSUPP);
	X3(E_REQUIRE_MEM_INSUFF, reqmem, 2048L, 4096L);
	X3(M_REQUIRE_MEM_OK, reqmem, 8192L, 4096L);
	X0(E_STRSUB_HUH);
	X0(M_PTY_PIPE);
	X2(E_EXEC_FAILED, 2, cmd);
	X1(M_FORK_SHELL_PATH, shell);
	X1(E_ENV_HIDDEN, hidden);
	X1(M_LOG_PATHNAME, logf);
	X1(E_LOG_PATHNAME_FAIL, logf);
	X0(E_ALIVE_USAGE);
	X1(E_ALIVE_BADHOST, host);
	X1(E_ALIVE_NOSOCKET, 5);
	X3(E_ALIVE_ATTEMPT, host, 5555, 3);
	X2(E_ALIVE_OK, host, 5555);
	X2(E_ALIVE_FAILED, host, 5555);
	X1(M_ENV, env);
	X1(E_UNLINK_OK, fulldir);
	X1(E_UNLINK_DIR, fulldir);
	X0(E_UNLINK_DOTDOT);
	X1(E_UNLINK_FAIL, fulldir);
	X0(E_UNLINK_NOPATH);
	X0(E_UNLINK_NOROOT);
	UNUSED(204);
	UNUSED(205);
	X4(E_UNLINK_XLATE, root, dirvar, root, dir);
	X2(M_EDIT_BEFORE, 12, &editsrch[1]);
	X1(M_EDIT_AFTER, editrepl);
	X1(E_STRSUB_OPEN_R, srcf);
	X1(E_STRSUB_OPEN_R_TMP, tmpf);
	X1(E_STRSUB_OPEN_W_TMP, tmpf);
	X2(E_STRSUB_OPEN_W, srcf, tmpf);
	X2(E_STRSUB_READ_ERROR, 5, srcf);
	X2(E_STRSUB_WRITE_ERROR, 28, tmpf);
	X3(M_STRSUB_REPLACING, &editsrch[1], editrepl, srcf);
	X1(M_EDIT_BEFORE, &editsrch[1]);
	X1(M_EDIT_AFTER, editrepl);
	X3(E_STRSUB_REWRITE_ERROR, 28, srcf, tmpf);
	X2(E_STRSUB_READ_TMP_ERROR, 5, tmpf);
	X2(M_STRSUB_SUMMARY, 100, 3);
	X0(E_MAIN_HOST_DENIED);
	X0(E_MAIN_BAD_ARG);
	X0(E_SSL_SE_INIT);
	X0(E_SSL_SE_BIOIN);
	X0(E_SSL_SE_BIOOUT);
	X0(E_SSL_SE_BADCA);
	X0(E_SSL_SE_BADCERT);
	X0(E_SSL_SE_BADPKEYFILE);
	X0(E_SSL_SE_BADPKEY);
	X0(E_SSL_SE_NOSSL);
	X0(E_SSL_SE_HANDSHAKE);
	X0(E_SSL_UNKNOWN);
	X2(E_SSL_UNVERIFIED_CERT, 5, sslcert);
	X1(M_SSL_STARTTLS, sslcipher);
	X1(M_SSL_CONSOLE_ISSUER, sslissuer);
	X1(M_SSL_CONSOLE_CERT, sslcert);
	X3(M_MONITOR_START, project, step, timestr);
	X3(M_MONITOR_END, project, step, timestr);
	X0(M_PTY_MASTER);
	X0(M_PTY_SLAVE);
	X3(M_CC_CALLS, cccaller, 42, cccallee);
	X2(M_CC_GENDIRDIFF, 5, 3);
	X0(E_CC_NO_PARSE_PATH);
	X1(M_ENV_SRC, srcf);
	X4(M_ENV_SET_INDIRECT, envvar, envvalue, envdotcmd, envdotcmdarg);
	X1(E_ENV_CMD_NO_FORMAT, envdotcmd);
	X0(E_ENV_SUBSTR_NO_ARGS);
	X0(E_ENV_SUBSTR_NO_SOURCE);
	X0(E_ENV_SUBSTR_BAD_START_INDEX);
	X0(E_ENV_SUBSTR_BAD_END_INDEX);
	X5(M_ENV_SUBSTR, envvar, envvalue, envprev, 3, 8);
	X7(M_ENV_REGEX, envvar, envvalue, editsrch, *editsrch, editrepl, *editsrch, envprev);
	X1(E_ENV_REGEX_BAD, &editsrch[1]);
	X2(M_ENV_SET, envvar, envvalue);
	X3(M_ENV_APPEND_PREPEND, envvar, envprev, envvalue);
	X1(M_ENV_ENCRYPTED, envvar);
	X1(M_ENV_HIDDEN, envvar);
	X1(M_ENV_CLEAR, envvar);
	X1(M_ENV_DELETE, envvar);
	X3(M_ENV_SKIPPED_ALREADY_SET, envvar, envprev, envvalue);
	X2(M_ENV_SET_NOT_ALREADY_SET, envvar, envvalue);
	X0(M_ENV_DEBUG);
	X1(E_MEM_REALLOC_UNKNOWN, ptr);
	X2(E_MEM_REALLOC_FREED, ptr, 42);
	X1(E_MEM_FREE_UNKNOWN, ptr);
	X2(E_MEM_FREE_FREED, ptr, 42);
	X1(M_MEM_REPORT_ALLOCS, 42L);
	X1(M_MEM_REPORT_REALLOCS, 42L);
	X1(M_MEM_REPORT_FREES, 42L);
	X1(M_MEM_REPORT_BYTES, 0x12345678L);
	X2(M_MEM_REPORT_UNFREED, ptr, 42);
	X0(E_AUTH_NO_USER);
	X0(M_WIN_SID_USER);
	X0(M_WIN_SID_GROUP);
	X0(M_WIN_SID_DOMAIN);
	X0(M_WIN_SID_ALIAS);
	X0(M_WIN_SID_WELL_KNOWN_GROUP);
	X0(M_WIN_SID_DELETED_ACCOUNT);
	X0(M_WIN_SID_INVALID);
	X0(M_WIN_SID_COMPUTER);
	X0(M_WIN_SID_UNKNOWN);
}


int main(int argc, char *argv[]) {
	int verbose = 1;
	int lang = 0;

	if (argc > 1 && strcmp(argv[1], "-q") == 0)
		verbose = 0;

	printf("locale: %s\n", platform_setlocale(""));

	while (langs[lang]) {
		if (verbose)
			printf("****************************************\n");
		printf("*** Using lang: [%s]\n", langs[lang]);
#ifdef WIN32
		printf("*** Using locale: [%s]\n",
				platform_setlocale(win_locale[lang]));
#endif
		if (verbose)
			printf("----------------------------------------\n");
		bfSetLang(langs[lang]);
		test_messages(verbose);
		++lang;
	}
	if (verbose)
		printf("========================================\n");

	return 0;
}

