/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#define MEM_C
#include "bf_agent.h"
#include <sys/types.h>

#define	TABLE_BLOCK_SIZE	4096

static unsigned int _allocs = 0;
static unsigned int _rallocs = 0;
static unsigned int _frees = 0;
static unsigned long _bytes = 0;
static int _memtbl_size = 0;
static struct _MEMTBL {
	int 	busy;
	unsigned long 	size;
	char	*addr;
} *_memtbl = NULL;

int _grow_memtbl(void) {
	char	*new;
	int 	new_size;

	new_size = _memtbl_size + TABLE_BLOCK_SIZE;
	if ((new = malloc(sizeof(struct _MEMTBL) * new_size)) == NULL)
		return 0;
	memset(new, 0, new_size * sizeof(struct _MEMTBL));
	if (_memtbl) {
		memcpy(new, _memtbl, _memtbl_size * sizeof(struct _MEMTBL));
		free(_memtbl);
	}
	_memtbl = (struct _MEMTBL *) new;
	_memtbl_size = new_size;
	return new_size;
}

void *bf_calloc(size_t nmemb, size_t size) {
	int 	i;

	if (!_memtbl)
		_grow_memtbl();
	for (i = 0 ; i < _memtbl_size ; i++) {
		if (!_memtbl[i].busy) {
			if ((_memtbl[i].addr = malloc(nmemb * size)) == NULL)
				return NULL;
			memset(_memtbl[i].addr, 0, nmemb * size);
			_memtbl[i].busy = 1;
			_memtbl[i].size = nmemb * size;
			++_allocs;
			_bytes += nmemb * size;
			return _memtbl[i].addr;
		}
	}
	if (!_grow_memtbl())
		return NULL;
	return bf_calloc(nmemb, size);
}

void *bf_malloc(size_t size) {
	return bf_calloc(1, size);
}

void *bf_realloc(void *p, size_t size) {
	char	*new;
	int 	i;

	if (!_memtbl)
		_grow_memtbl();

	for (i = 0 ; i < _memtbl_size ; i++)
		if (_memtbl[i].addr == p)
			break;
	if (i == _memtbl_size) {
		if (bf_debug) {
			format_message(bf_debug, bfCat(E_MEM_REALLOC_UNKNOWN), p);
			DebugBlock(bf_debug);
		}
		return NULL;
	}
	if (!_memtbl[i].busy) {
		if (bf_debug) {
			format_message(bf_debug, bfCat(E_MEM_REALLOC_FREED),
					_memtbl[i].addr, _memtbl[i].size);
			DebugBlock(bf_debug);
		}
		return NULL;
	}
	++_rallocs;
	_bytes += size;
	new = bf_calloc(1, size);
	memcpy(new, p, _memtbl[i].size);
	free(_memtbl[i].addr);
	_memtbl[i].addr = new;
	_memtbl[i].size = size;
	return new;
}

void _memreport(void) {
	const char *unfreed;
	int 	i;

	if (!bf_debug)
		return;

	format_message(bf_debug, bfCat(M_MEM_REPORT_ALLOCS), _allocs);
	DebugBlock(bf_debug);
	format_message(bf_debug, bfCat(M_MEM_REPORT_REALLOCS), _rallocs);
	DebugBlock(bf_debug);
	format_message(bf_debug, bfCat(M_MEM_REPORT_FREES), _frees);
	DebugBlock(bf_debug);
	format_message(bf_debug, bfCat(M_MEM_REPORT_BYTES), _bytes);
	DebugBlock(bf_debug);

	unfreed = bfCat(M_MEM_REPORT_UNFREED);
	for (i = 0 ; i < _memtbl_size ; i++) {
		if (_memtbl[i].busy) {
			format_message(bf_debug, unfreed, _memtbl[i].addr, _memtbl[i].size);
			DebugBlock(bf_debug);
		}
	}
}

void bf_free(void *p) {
	int 	i;

	if (!p) {
		_memreport();
		return;
	}
	for (i = 0 ; i < _memtbl_size ; i++) {
		if (p == _memtbl[i].addr) {
			if (_memtbl[i].busy) {
				free(_memtbl[i].addr);
				_memtbl[i].busy = 0;
				++_frees;
			}
			else {
				if (bf_debug) {
					format_message(bf_debug, bfCat(E_MEM_FREE_FREED),
							_memtbl[i].addr, _memtbl[i].size);
					DebugBlock(bf_debug);
				}
			}
			return;
		}
	}
	if (bf_debug) {
		format_message(bf_debug, bfCat(E_MEM_FREE_UNKNOWN), p);
		DebugBlock(bf_debug);
	}
}

