/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int FuncMkdir(void)
{
	char	buf[BFBUFSIZ], msg[BFBUFSIZ], *root, *odir, *dir, *p;
	char	*ccroot, *timeout, *userdata;
	int 	mode;
	int 	mapfail;
	size_t	userdata_size;

	/*  locate the root element */
	if (NULL == (root = request_get(TAG_ROOT))) {
		ResultBlock(E_MKDIR_NOROOT, -1, bfCat(E_MKDIR_NOROOT));
		return 0;
	}
	root = strdup(root);

	/*  locate the dir element */
	if (NULL == (odir = request_get(TAG_DIR))) {
		ResultBlock(E_MKDIR_NODIR, -1, bfCat(E_MKDIR_NODIR));
		free(root);
		return 0;
	}

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the client timeout element */
	if (NULL == (timeout = request_get(TAG_TIMEOUT)))
		timeout = "0";

	/*  locate the silent element */
	verbose = !request_has(TAG_SILENT);

	StoreUserData(userdata, userdata_size);

	DumpEnv();

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		ResultBlock(E_EXEC_MAPFAIL, -1, bfCat(E_EXEC_MAPFAIL));
		free(root);
		FreeUserData();
		return 0;
	}

	ClobberPasswords();

	ccroot = NULL;
	if (clearcase_view) {
		ccroot = calloc(1, BFBUFSIZ);
		format_message(ccroot, "%s/%s/%s", this_view(), clearcase_view, root);
		free(root);
		root = ccroot;
		ccroot = NULL;
	}

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/*  assemble and clean the final pathname. clean the
	 *  pieces so we can use them individually later. */
	CleanPath(root);
	CleanPath(odir);

	root = env_expand_var((p = root), 0, 0);
	free(p);
	dir = env_expand_var(odir, 0, 0);
	CleanPath(dir);
	if (strcmp(dir, odir)) {
		format_message(msg, bfCat(E_MKDIR_XLATE), root, odir, root, dir);
		CleanPath(msg);
		if (verbose)
			DataBlock(msg, NULL);
	}

	format_message(buf, "%s/%s", root, dir);
	CleanPath(buf);

	/*  test to see if it exists already */
	if (isDir(buf)) {
		platform_unmap_drives();
		format_message(msg, bfCat(E_MKDIR_EXISTS), buf);
		ResultBlock(0, -1, msg);
		free(dir);
		if (ccroot)
			free(ccroot);
		free(root);
		FreeUserData();
		return 0;
	}

	/*  locate the mode element. if it wasn't supplied, set a default. */
	mode = 0755;
	p = request_get(TAG_MODE);
	if (p != NULL)
		sscanf(p, "%o", (unsigned int *)&mode);

	BFCreateDirectory(root, dir, mode);

	/*  now just test to see if the dir is there. if it is, the 
	 *  above code musta worked. */
	if (isDir(buf)) {
		format_message(msg, bfCat(E_MKDIR_OK), buf);
		ResultBlock(0, -1, msg);
	}
	else {
		format_message(msg, bfCat(E_MKDIR_FAIL), buf);
		ResultBlock(E_MKDIR_FAIL, -1, msg);
	}

	platform_unmap_drives();
	free(dir);
	if (ccroot)
		free(ccroot);
	free(root);
	FreeUserData();
	return 0;
}

