/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

#define PLATFORM_C
#define CONVERT_USE_STATIC_BUFFER
#include "bf_agent.h"
#define hRemote ofd

int bf_openpty(void);
int utf8_is_native = -1;

extern  int     ifd;
extern  int     ofd;
extern  int     errno;
extern  char ** environ;

static  int     alldone; /* used for monitoring child process */
static  int     child_pgrp;
static  uid_t   original_uid;
static  uid_t   original_gid;
static  FILE *  monitor_f   = NULL;
static  int     monitor_id  = 0;
static  char    scbuf[BFBUFSIZ];
static  char *  signal_catcher = NULL;

static  char *  iconv_native_encoding = NULL;
static  iconv_t cd_utf8_to_native = ICONV_INVALID;
static  iconv_t cd_native_to_utf8 = ICONV_INVALID;

int     trapmode = 0;

extern bf_command_state_t command_state;

static const char *encoding_aliases[] = {
	"646",       "ASCII", "US-ASCII", NULL,
	"CP20866",   "KOI8-R", "KOI8-U", "KOI8-UR", NULL,
	"CP20936",   "GB2312", "GB18030", NULL,
	"CP21866",   "KOI8-U", "KOI8-UR", "KOI8-R", NULL,
	"CP28591",   "ISO-8859-1", "ISO-8859-15", NULL,
	"CP28592",   "ISO-8859-2", NULL,
	"CP28593",   "ISO-8859-3", NULL,
	"CP28594",   "ISO-8859-4", NULL,
	"CP28595",   "ISO-8859-5", NULL,
	"CP28596",   "ISO-8859-6", NULL,
	"CP28597",   "ISO-8859-7", NULL,
	"CP28598",   "ISO-8859-8", NULL,
	"CP28599",   "ISO-8859-9", NULL,
	"CP28605",   "ISO-8859-15", "ISO-8859-1", NULL,
	"CP38598",   "ISO-8859-8", NULL,
	"CP51932",   "EUC-JP", NULL,
	"CP51936",   "GB2312", NULL,
	"CP51949",   "EUC-KR", NULL,
	"CP51950",   "EUC-TW", NULL,
	"CP54936",   "GB18030", "GBK", NULL,
	"CP65000",   "UTF-7", NULL,
	"CP65001",   "UTF-8", NULL,
	"JIS7",      "SHIFTJIS", NULL,
	"JOHAB",     "CP1361", NULL,
	"PT154",     "CP1251", NULL,
	"WINSAMI2",  "CP1252", NULL,
	NULL
};

static const char **get_encoding_aliases(const char *encoding) {
	const char **mark = encoding_aliases;
	char *enc = malloc(strlen(encoding) + 1);
	unsigned const char *p = (unsigned const char *)encoding;
	unsigned char *q = (unsigned char *)enc;

	while (*p) {
		if (*p != ' ' && *p != '-')
			*q++ = (unsigned char)toupper(*p);
		++p;
	}
	while (NULL != *mark) {
		if (0 == strcmp(*mark++, enc))
			break;
		while (NULL != *mark)
			++mark;
		++mark;
	}
	free(enc);
	return mark;
}

/* Returns 1 if encoding matches /^UTF-?8/i or 0 otherwise. */
static int is_encoding_utf8(const char *encoding) {
	if (NULL == encoding) return 0;
	if (*encoding != 'U' && *encoding != 'u') return 0;
	++encoding;
	if (*encoding != 'T' && *encoding != 't') return 0;
	++encoding;
	if (*encoding != 'F' && *encoding != 'f') return 0;
	++encoding;
	if (*encoding == '-') ++encoding;
	if (*encoding++ != '8') return 0;
	if (*encoding != '\0') return 0;
	return 1;
}

static int bf_iconv_open_using(const char *encoding) {
	cd_utf8_to_native = iconv_open(encoding, "UTF-8");
	if (cd_utf8_to_native == ICONV_INVALID) {
		if (bf_debug) {
			format_message(bf_debug, "bf_iconv_open_using: UTF-8 -> %s: %d", encoding, errno);
			DebugBlock(bf_debug);
		}
		return 0;
	}

	cd_native_to_utf8 = iconv_open("UTF-8", encoding);
	if (cd_native_to_utf8 == ICONV_INVALID) {
		if (bf_debug) {
			format_message(bf_debug, "bf_iconv_open_using: %s -> UTF-8: %d", encoding, errno);
			DebugBlock(bf_debug);
		}
		iconv_close(cd_utf8_to_native);
		cd_utf8_to_native = ICONV_INVALID;
		return 0;
	}

	iconv_native_encoding = strdup(encoding);
	return 1;
}

static void bf_iconv_open(void) {
	const char *encoding, **aliases;

	if (NULL != iconv_native_encoding)
		return;

	if (BFCONF_EXISTS(disable_transcode)) {
		if (bf_debug)
			DebugBlock("bf_iconv_open: disable_transcode");
		utf8_is_native = 1;
		return;
	}

	encoding = nl_langinfo(CODESET);
	if (NULL == encoding)
		encoding = "US-ASCII";

	utf8_is_native = is_encoding_utf8(encoding);
	if (utf8_is_native) {
		if (bf_debug)
			DebugBlock("bf_iconv_open: UTF-8");
		iconv_native_encoding = strdup(encoding);
		return;
	}

	if (bf_iconv_open_using(encoding)) {
		if (bf_debug) {
			format_message(bf_debug, "bf_iconv_open: encoding=%s", encoding);
			DebugBlock(bf_debug);
		}
		return;
	}

	aliases = get_encoding_aliases(encoding);
	while (NULL != *aliases) {
		if (bf_iconv_open_using(*aliases))
			return;
		++aliases;
	}

	if (bf_iconv_open_using("US-ASCII") || bf_iconv_open_using("ASCII")) {
		if (bf_debug)
			DebugBlock("bf_iconv_open: falling back on ASCII");
		return;
	}

	if (bf_debug)
		DebugBlock("bf_iconv_open: falling back on UTF-8");
	utf8_is_native = 1;
	iconv_native_encoding = strdup("UTF-8");
	return;
}

static void bf_iconv_close(void) {
	if (NULL == iconv_native_encoding)
		return;

	free(iconv_native_encoding);
	iconv_native_encoding = NULL;
	utf8_is_native = -1;
	if (cd_utf8_to_native != ICONV_INVALID) {
		iconv_close(cd_utf8_to_native);
		cd_utf8_to_native = ICONV_INVALID;
	}
	if (cd_native_to_utf8 != ICONV_INVALID) {
		iconv_close(cd_native_to_utf8);
		cd_native_to_utf8 = ICONV_INVALID;
	}
	if (bf_debug)
		DebugBlock("bf_iconv_close");
}

static void ignore_signal(int sig) {
	if (bf_debug) {
		printf("[%d] %s: signal %d: ignored\n",
				(int)getpid(),
				signal_catcher,
				sig);
	}
	signal(sig, ignore_signal);
}

static void fatal_signal(int sig) {
	if (bf_debug) {
		printf("[%d] %s: signal %d: exiting\n",
				(int)getpid(),
				signal_catcher,
				sig);
	}
	exit(sig);
}

static void reap_zombies(int sig) {
	int status;
	pid_t pid;
	while ( (pid = waitpid(-1, &status, WNOHANG)) > 0 ) {
		if (bf_debug) {
			printf("[%d] %s: child %d exited with status %d\n",
				(int)getpid(),
				signal_catcher,
				(int)pid,
				status );
		}
	}
	signal(SIGCHLD, reap_zombies);
}


char *this_view(void) {
	static char _viewroot[256];
	char	*p;

	p = bfconf_get("ccviewroot", "/view");
	strcpy(_viewroot, p);
	free(p);
	return _viewroot;
}


/* function that gets call for all output to the console */
/* used to update a debug window if one exists, implemented for windows only currently */
void MonitorOutput(const char *str) {
	if (monitor_f) {
		fprintf(monitor_f, "%d %s\n", monitor_id, str);
		fflush(monitor_f);
	}
}
void MonitorOpen(void) {
	char	*p, *proj, *step;
	char	info[BFBUFSIZ];
	time_t	now;

	MonitorClose();
	if ((p = bfconf_get("monitor_pipe", NULL)) != NULL && access(p, 0) == 0) {
		monitor_f = fopen(p, "w");
		free(p);
	}
	monitor_id = getpid();
	if (NULL == (proj = platform_getenv("BF_PROJECTNAME")))
		proj = "Unknown";
	if (NULL == (step = platform_getenv("BF_STEPNAME")))
		step = "Unknown";
	time(&now);
	format_message(info, bfCat(M_MONITOR_START), proj, step, ctime(&now));
	MonitorOutput(info);
}

void MonitorClose(void) {
	char *proj, *step, info[BFBUFSIZ];
	time_t	now;

	if (monitor_f) {
		if (NULL == (proj = platform_getenv("BF_PROJECTNAME")))
			proj = "Unknown";
		if (NULL == (step = platform_getenv("BF_STEPNAME")))
			step = "Unknown";
		time(&now);
		format_message(info, bfCat(M_MONITOR_END), proj, step, ctime(&now));
		MonitorOutput(info);
		fclose(monitor_f);
	}
	monitor_f = NULL;
}

char **getBFenviron(void) {
	return environ;
}

void freeBFenviron(char **p) {
	/* On Windows, getBFenviron gets a copy of the environment which
	 * must later be freed.  On Unix, it returns a live reference to
	 * the actual environment, so this is a no-op on those platforms.
	 */
}

/*
 * spew the current environment to the engine
 *
 */
void DumpEnv(void) {
	const char *msg;
	char	**env, **env_ptr;
	char	*buf;
	int 	bsize;

	if (!verbose)
		return;
#ifdef HAVE_SSL
	report_SSL();
#endif
	buf = malloc(BUFSIZ);
	bsize = BUFSIZ;
	msg = bfCat(M_ENV);
	for (env = env_ptr = getBFenviron(); *env; ++env) {
		if (strlen(*env) + 6 > (unsigned int) bsize) {
			buf = realloc(buf, strlen(*env) + BUFSIZ);
			bsize = strlen(*env) + BUFSIZ;
		}
		format_message(buf, msg, *env);
		if (env_is_hidden(*env))
			strcpy(strchr(buf, '='), "=*****");
		DataBlock(buf, NULL);
	}
	freeBFenviron(env_ptr);
	free(buf);
}

static RETSIGTYPE client_timeout_handler(int signum) {
	kill(child_pgrp, SIGTERM);
	sleep(2);
	kill(child_pgrp, SIGKILL);
	alldone = 1;
}

/* #define PLATFORM_RELAY_DEBUG */

#ifdef PLATFORM_RELAY_DEBUG
#define PRD(x) x
#else
#define PRD(x) /**/
#endif

#define RCO_NATIVE_BUFSIZ (BFBUFSIZ << 1)
#define RCO_UTF8_BUFSIZ (BFBUFSIZ << 2)
static char *rco_native_buf = NULL;
static int rco_native_len = 0;
static char *rco_utf8_buf = NULL;
static int rco_utf8_len = 0;
static int relay_command_output(int fd, int flush_all)
{
	int     bytes, len, got_data = 0;
	int     saved = -1;
	char	*p;

	if (rco_utf8_buf == NULL)
		rco_utf8_buf = malloc(RCO_UTF8_BUFSIZ);

	if (platform_utf8_is_native()) {
		/* Read UTF-8 data */
		while ((len = read(fd,
						&rco_utf8_buf[rco_utf8_len],
						RCO_UTF8_BUFSIZ - rco_utf8_len)) > 0)
		{
			got_data = 1;
			rco_utf8_len += len;
		}
	} else {
		if (NULL == rco_native_buf)
			rco_native_buf = malloc(RCO_NATIVE_BUFSIZ);

		/* Read non-UTF-8 data, transcoding whenever we reach a good stopping point. */
		while ((len = read(fd,
						&rco_native_buf[rco_native_len],
						RCO_NATIVE_BUFSIZ - rco_native_len)) > 0)
		{
			got_data = 1;
			rco_native_len += len;
			if ((len = convert_native_find_boundary(rco_native_buf, rco_native_len)) <= 0) {
				/* No convenient boundary.  If there is still room in the
				 * buffer, then wait for more data.  Otherwise, give up and
				 * transcode whatever we have.
				 */
				if (rco_native_len < RCO_NATIVE_BUFSIZ)
					continue;
				len = BFBUFSIZ;
			}

			/* Transcode! */
			bytes = convert_native_to_utf8(
					rco_native_buf,
					len,
					&rco_utf8_buf[rco_utf8_len],
					RCO_UTF8_BUFSIZ - rco_utf8_len);
			PRD(printf("relay_command_output: native len=%d -> %d\n", len, bytes));
			if (bytes > 0) {
				/* Adjust the buffers */
				rco_native_len -= len;
				memmove(rco_native_buf, &rco_native_buf[len], rco_native_len);
				rco_utf8_len += bytes;
			}
			break;
		}
	}

	/* Look for completed lines.  If the buffer exceeds BFBUFSIZ, then go
		* ahead and dump it that much data.
		*/
	while (rco_utf8_len > 0) {
		PRD(printf("relay_command_output: utf8 len=%d\n", (int)rco_utf8_len));
		p = memchr(rco_utf8_buf, '\n', rco_utf8_len);
		if (NULL == p) {
			if (rco_utf8_len < BFBUFSIZ)
				break;
			PRD(printf("relay_command_output: splitting long line\n"));
			p = &rco_utf8_buf[BFBUFSIZ];
			saved = *p & 0xFF;
		}

		/* We discard newlines and carriage returns.  This is horrible and
		 * we should change the behavior when we get a chance.
		 */
		*p = '\0';
		if (p > rco_utf8_buf && p[-1] == '\r')
			p[-1] = '\0';

		if (!trapmode && !DataBlock(rco_utf8_buf, NULL)) {
			client_timeout_handler(0);
			return got_data;
		}
		if (saved >= 0) {
			*p = (char)saved;
			saved = -1;
		}
		else {
			++p;
		}
		rco_utf8_len -= p - rco_utf8_buf;
		memmove(rco_utf8_buf, p, rco_utf8_len);
	}

	if (flush_all) {
		if (!platform_utf8_is_native() && rco_native_len > 0) {
			PRD(printf("relay_command_output: final flush native=%d\n", rco_native_len));
			bytes = convert_native_to_utf8(
					rco_native_buf,
					len,
					&rco_utf8_buf[rco_utf8_len],
					RCO_UTF8_BUFSIZ - rco_utf8_len);
			if (bytes > 0)
				rco_utf8_len += bytes;
			rco_native_len = 0;
		}
		if (rco_utf8_len > 0) {
			PRD(printf("relay_command_output: final flush utf8=%d\n", rco_utf8_len));
			rco_utf8_buf[rco_utf8_len] = '\0';
			if (!trapmode && !DataBlock(rco_utf8_buf, NULL))
				client_timeout_handler(0);
			rco_utf8_len = 0;
		}
	}

	return got_data;
}

int LoadSources(int fd, char *list) {
	char    buf[BFBUFSIZ];
	char    *p;
	char    *bf_rptr=NULL;
	char    *bf_sptr=NULL;
	FILE    *f;

	p = bf_strtok(list, "\n", &bf_rptr, &bf_sptr);
	while (p) {
		if ((f = fopen(p, "r")) == NULL) {
			if (BFCONF_EXISTS(fail_if_dot_source_fails)) {
				format_message(buf, bfCat(E_EXEC_COPYFAIL), p, errno);
				DataBlock(buf, NULL);
				return E_EXEC_COPY;
			}
		}
		else {
			format_message(buf, bfCat(E_EXEC_COPY), p);
			DataBlock(buf, NULL);
			while (fgets(buf, BFBUFSIZ, f) != NULL)
				write(fd, buf, strlen(buf));
			fclose(f);
		}
		p = bf_strtok(NULL, "\n", &bf_rptr, &bf_sptr);
	}
	return 0;
}


static int write_input(int fd, char *buf, int len, char **convert)
{
	if (*convert) {
		int max = BFBUFSIZ << 1;
		if (len > BFBUFSIZ)
			*convert = realloc(*convert, max = len << 1);
		len = convert_utf8_to_native(buf, len, *convert, max);
		return write(fd, *convert, len) == len;
	}
	return write(fd, buf, len) == len;
}

/* shell is the exec absolute path to the shell */
/* sharg is argv[0] to the process */
static int buildInput(char **namebuf, char *preload, char *cmd, char *indata, char *cdir, char **shell, char **sharg) {
	char 	*p, *sh = NULL, *buf = NULL, *convert = NULL, *final_cmd = NULL;
	int 	fd, custom_shell, len, no_preparse;


	/* build a script to execute */
	buf = malloc(BFBUFSIZ);
	if (!platform_utf8_is_native())
		convert = malloc(BFBUFSIZ << 1);

	if ((fd = platform_tempfd(namebuf)) < 0) {
		format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, *namebuf);
		DataBlock(buf, NULL);
		goto done;
	}
	platform_chmod(*namebuf, 0755);

	if (clearcase_view) {
		format_message(buf, bfCat(E_EXEC_CCVIEW), clearcase_view);
		DataBlock(buf, NULL);
	}

	/* if the cmd string starts with #!, then use that shell */
	sh = strdup(cmd); /* work on a copy so we don't clobber the original cmd string */
	if (strncmp(sh, "#!", 2) == 0 && (p = strchr(sh, '\n')) != NULL) {
		*p++ = 0;
		final_cmd = strdup(p);			/* actual command is the remainder of the string */
		strcpy(sh, &sh[2]);	/* eat the #! */
		free(*shell); *shell = strdup(sh);		/* set the actual shell path */
		free(*sharg);
		if ((p = strrchr(sh, '/')) != NULL)	/* set arg0 to the base shell pathname */
			++p;
		else
			p = sh;
		*sharg = strdup(p);
		custom_shell = 1;
	}
#ifdef HAVE_REXX
	/* Some systems support REXX natively using a REXX-style comment instead
	 * of a #! line.  This will support that invocation style directly, as
	 * on some systems the native support also means that there isn't an
	 * interpreter available for it.  z/OS springs to mind.
	 */
	else if (strncmp(sh, "/*", 2) == 0 && (p = strchr(sh, '\n')) != NULL
			&& (strstr(sh, "REXX") < p ||
				strstr(sh, "rexx") < p ||
				strstr(sh, "Rexx") < p))
	{
		custom_shell = 1;
		free(*shell);  *shell = strdup(*namebuf);	/* set the actual shell path */
		free(*sharg);  *sharg = NULL;
		final_cmd = strdup(cmd);
	}
#endif /* HAVE_REXX */
	else {
		custom_shell = 0;
		final_cmd = strdup(cmd);
	}
	free(sh);

	/* write out the shell prefix */
	if (*sharg) {
		len = format_message(buf, "#!%s\n", *shell);
		if (!write_input(fd, buf, len, &convert)) {
			format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, *namebuf);
			DataBlock(buf, NULL);
			fd = -1;
			goto done;
		}
	}

	/* write out a cd to the command directory. the process will be sitting in $HOME */
	/* so that profile processing is correct.  This is not done if the shell has been
	 * changed, because we can't be sure it is a shell at all.  It could be Perl, awk,
	 * REXX, etc.
	 */
	if (!custom_shell) {
		len = format_message(buf, "cd %s\n", cdir);
		if (!write_input(fd, buf, len, &convert)) {
			format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, *namebuf);
			DataBlock(buf, NULL);
			fd = -1;
			goto done;
		}
	}

	if (preload && *preload != '\0' && LoadSources(fd, preload) != 0) {
		close(fd);
		unlink(*namebuf);
		fd = -1;
		goto done;
	}

	no_preparse = BFCONF_EXISTS(no_preparse_command);
	if (!no_preparse) {
		p = platform_getenv("_NO_PREPARSE_COMMAND");
		if (p) {
			no_preparse = 1;
			platform_freeenv(p);
		}
	}

	if (no_preparse)
		sh = strdup(final_cmd);
	else
		sh = env_expand_hidden(final_cmd, 1, 1);

	/* Make sure that the command ends with a newline.
	 * We also grow the buffers as needed to hold the command.
	 */
	len = strlen(sh) + 1;
	if (len > BFBUFSIZ)
		buf = realloc(buf, len);

	memcpy(buf, sh, len-1);
	buf[len-1] = '\n';
	if (!write_input(fd, buf, len, &convert)) {
		format_message(buf, bfCat(E_EXEC_SCRIPTFAIL), errno, *namebuf);
		DataBlock(buf, NULL);
		fd = -1;
	}

done:
	free(convert);
	free(buf);
	free(sh);
	free(final_cmd);
	if (fd < 0) {
		free(*namebuf);
		*namebuf = NULL;
	} else {
		lseek(fd, 0, SEEK_SET);
	}
	return fd;
}


int run_process(
		char *dir,
		char *cmd,
		int timeout,
		char *indata,
		char *srclist,
		char *startmsg,
		int *timeout_result,
		int close_pty)
{
	bf_command_state_t *state = &command_state;
	int 	pid, pstatus, fd = -1, warnout, warned, ret;
	int 	nologonshell, shellargmode;
	int		sync_pipe[2];
	time_t	out_of_time;
	char	*buf = NULL, *p, *shell = NULL, *shellpath = NULL, *home = NULL, *logname = NULL;
	char	*namebuf = NULL, *shellflag = NULL;
	struct	passwd	*pwd;

	buf = calloc(1, BFBUFSIZ << 1);

	warnout = 0;
	if ((p = platform_getenv("_WARNOUT")) != NULL) {
		warnout = atoi(p);
		platform_freeenv(p);
	}

	/* get the user's shell */
	nologonshell = BFCONF_EXISTS(nologonshell);
	if ((pwd = getpwuid(getuid())) != NULL) {
		if ((shellpath = bfconf_get("shell", NULL)) == NULL) {
			shellpath = strdup(pwd->pw_shell);
			if ((p = strrchr(shellpath, '/')) != NULL) {
				shell = malloc(strlen(++p) + 2);
				format_message(shell, "%s%s", (nologonshell) ? "" : "-", p);
			}
			else {
				shell = strdup(shellpath);
			}
		}
		else {
			shell = strdup(shellpath);
		}
		home = strdup(pwd->pw_dir);
		logname = strdup(pwd->pw_name);
		endpwent();
	}
	else {
		shell = malloc(4);
		format_message(shell, "%ssh", (nologonshell) ? "" : "-");
		shellpath = bfconf_get("shell", "/bin/sh");
		home = strdup("/tmp");
		logname = strdup("buildforge");
	}

	if ((fd = buildInput(&namebuf, srclist, cmd, indata, dir, &shellpath, &shell)) < 0) {
		ret = errno;
		goto done;
	}

	format_message(buf, bfCat(M_FORK_SHELL_PATH), shellpath);
	DataBlock(buf, NULL);

	/* sync_pipe[0] is reader, sync_pipe[1] is writer ... */
	if (pipe(sync_pipe) != 0) {
			/* pipe failed - system error */
			format_message(buf, bfCat(E_EXEC_EXECFAIL), errno);
			DataBlock(buf, NULL);
			unlink(namebuf);
			close(fd);
			if (close_pty) {
				close(state->sfd);
				close(state->mfd);
			}
			ret = errno;
			goto done;
	}

	/* fork off a child process */
	switch (pid = fork()) {
		case -1:	/* fork failed - system error */
			format_message(buf, bfCat(E_EXEC_EXECFAIL), errno);
			DataBlock(buf, NULL);
			unlink(namebuf);
			close(fd);
			if (close_pty) {
				close(state->sfd);
				close(state->mfd);
			}
			close(sync_pipe[0]);
			close(sync_pipe[1]);
			ret = errno;
			break;

		case 0: 	/* this is the child process */
			/* we don't care about the alloc'd pointer coming back, this is the child and we're */
			/* just going to exit shortly */

			/* Block until the parent closes sync_pipe[1] */
			close(sync_pipe[1]);
			read(sync_pipe[0], buf, 1);
			close(sync_pipe[0]);

			setpgid(getpid(), getpid());
			close(state->mfd);
			platform_setenv("HOME", home, 1);
			platform_setenv("LOGNAME", logname, 1);
			platform_setenv("SHELL", shellpath, 1);
			{
				const char *locale = platform_setlocale(NULL);
				if (NULL != locale)
					setenv("LANG", locale, 1);
			}

			if (clearcase_view) {
				close(fd);
				dup2(state->sfd, 0);
				dup2(state->sfd, 1);
				dup2(state->sfd, 2);
				close(state->sfd);
				platform_chdir(home); /* cd to $HOME to properly handle relative items in .profiles */
				if (cc_audit) {
					format_message(buf, "clearaudit -c %s", namebuf);
					execlp("cleartool", "cleartool", "setview", "-login", "-exec", buf, clearcase_view, NULL);
				}
				else
					execlp("cleartool", "cleartool", "setview", "-login", "-exec", namebuf, clearcase_view, NULL);
				printf(bfCat(E_CTOOL_EXECFAIL), clearcase_view);
				exit(errno);
			}
			else {
				shellflag = bfconf_get("shellflag", NULL);
				shellargmode = BFCONF_EXISTS(shellarg);
				if (shell && !shellargmode && dup2(fd, 0) < 0) {
					printf(bfCat(E_FORK_STREAMFAIL), errno);
					close(1);
					close(2);
					close(state->sfd);
					close(fd);
					exit(errno);
				}
				close(fd);
				dup2(state->sfd, 1);
				dup2(state->sfd, 2);
				if (shell && !shellargmode)
					close(state->sfd);
				platform_chdir(home); /* cd to $HOME to properly handle relative items in .profiles */
				if (!shell) {
					/* No shell is being used.  For example, REXX under z/OS. */
					platform_chdir(dir);
					dup2(state->sfd, 0);
					close(state->sfd);
					execl(shellpath, shellpath, NULL);	/* /tmp/script */
				} else if (shellargmode) {
					dup2(state->sfd, 0);
					close(state->sfd);
					if (shellflag && *shellflag != '\0') {
						execl(shellpath, shell, shellflag, namebuf, NULL); /* /bin/csh -f /tmp/script */
					} else {
						execl(shellpath, shell, namebuf, NULL); /* /bin/sh /tmp/script */
					}
				}
				else {
					if (shellflag && *shellflag != '\0') {
						execlp(shellpath, shell, shellflag, NULL); /* /bin/csh -f < /tmp/script */
					} else {
						execlp(shellpath, shell, NULL); /* /bin/sh < /tmp/script */
					}
				}
				printf(bfCat(E_EXEC_FAILED), errno, shellpath);
				exit(errno);
			}

		default:	/* this is the parent process */
			close(sync_pipe[0]);  /* close the reader; we don't need it */
			close(fd);
			if (close_pty)
				close(state->sfd);
			if (startmsg)
				DataBlock(startmsg, NULL);

			signal(SIGTERM, client_timeout_handler);
			signal(SIGPIPE, client_timeout_handler);
			signal(SIGINT, SIG_IGN);
			signal(SIGQUIT, SIG_IGN);
			signal(SIGHUP, SIG_IGN);
			signal(SIGCHLD, SIG_DFL);

			child_pgrp = -pid;

			fcntl(state->mfd, F_SETFL, O_NONBLOCK);

			/* unblock the child */
			close(sync_pipe[1]);

			/* read the output from the child. */
			out_of_time = time(NULL) + timeout + 1;
			if (!timeout)
				warnout = 0;
			alldone = 0;
			warned = 0;
			/* keep reading while the child is alive and we haven't timed out */
			OutputCache(CACHE_ON);
			while (!alldone) {
				/* check for timeout */
				if (timeout && time(NULL) > out_of_time) {
					alldone = 1;
					continue;
				}

				/* check for warnout */
				if (warnout && !warned && time(NULL) > (out_of_time - warnout)) {
					warned = 1;
					format_message(buf, bfCat(E_EXEC_WARNOUT), warnout);
					OutputCache(CACHE_OFF);
					DataBlock(buf, NULL);
					OutputCache(CACHE_ON);
					continue;
				}

				/* check for child exit */
				if (waitpid(pid, &pstatus, WNOHANG) > 0) {
					alldone = 1;
					continue;
				}

				if (relay_command_output(state->mfd, 0)) {
					/* reset the timeout counter because we read *some* data */
					out_of_time = time(NULL) + timeout + 2;
					warned = 0;
				}

				usleep(2500);
				if (!trapmode && !HeartBeat())
					client_timeout_handler(0);
			}
			OutputCache(CACHE_OFF);

			/* flush any remaining data that might be left over (never got a newline) */
			relay_command_output(state->mfd, 1);

			/* we get here if we didn't get any data and should time out */
			if (timeout && time(NULL) > out_of_time) {
				if (timeout_result)
					*timeout_result = timeout;
				kill(child_pgrp, SIGTERM);
				sleep(2);
				kill(child_pgrp, SIGKILL);
				kill(pid, 9);
				sleep(1);
				waitpid(pid, &pstatus, WNOHANG);
				ret = 999;
			} else {
				ret = WEXITSTATUS(pstatus);
			}
			if (close_pty)
				close(state->mfd);
			break;
	}

done:
	if (namebuf) {
		cleanMyCruft(namebuf);
		free(namebuf);
	}
	free(buf);
	free(shell);
	free(shellpath);
	free(home);
	free(logname);
	return ret;
}

/*
 * function to remove a directory and all its contents.
 * this is, oddly, very platform specific. this common
 * module is written for most unixes.
 */
#if HAVE_FTW_H
#include <ftw.h>
#undef HAVE_FTS_H
#endif

#ifdef FTW_DEPTH
int RemoveDirectoryFunction(
		const char *path,
		const struct stat *sb,
		int flags,
		struct FTW *f)
{
	char	buf[BFBUFSIZ];

	if (platform_unlink(path) && platform_rmdir(path)) {
		format_message(buf, bfCat(E_RMDIR_RMERR), errno, path);
		DataBlock(buf, NULL);
	}
	return 0;
}

void BFRemoveDirectory(char *path)
{
	nftw(path, RemoveDirectoryFunction, 50, FTW_DEPTH|FTW_PHYS);
}
#else
void BFRemoveDirectory(char *path)
{
	pid_t	pid;
	int 	pstatus;

	switch ((pid = fork())) {
		case -1: /* fork failed */
			break;

		case 0:
			execl("/bin/rm", "rm", "-rf", path, NULL);
			exit(errno);

		default:
			waitpid(pid, &pstatus, 0);
			break;
	}
}
#endif


#ifdef HAVE_FTS_H
#include <fts.h>

int fts_null_cmp(const FTSENT **a, const FTSENT **b) {
	return 0;
}

#define FTW_F		FTS_F
#define FTW_D		FTS_D
#define FTW_SL		FTS_SL
#define FTW_NS		FTS_NS

#define FTS_OPTS	(FTS_F|FTS_SL|FTS_SLNONE|FTS_NS|FTS_D)

int ftw(
		const char *path,
		int (*func)(const char *, const struct stat *, int),
		int depth)
{
	FTS				*fts;
	FTSENT			*ent;
	struct stat		sb;
	char			*paths[2], *p;
	int 			flags;

	paths[0] = p = strdup(path);
	paths[1] = NULL;

	fts = fts_open(paths, FTS_NOCHDIR|FTS_NOSTAT|FTS_PHYSICAL, fts_null_cmp);
	if (fts == NULL) {
		free(p);
		return errno;
	}

	while ((ent = fts_read(fts)) != NULL) {
		if (ent->fts_info & FTS_OPTS) {
			if (lstat(ent->fts_accpath, &sb) == 0)
				func(ent->fts_accpath, &sb, ent->fts_info & FTS_OPTS);
		}
	}
	fts_close(fts);
	free(p);
	return 0;
}
#endif


int BFGetPath(char *path, int (*Func)(const char *, const struct stat *, int)) {
#ifdef HAVE_FTS_H
	return ftw(path, Func, 50);
#else
	return ftw(path, Func, 50) >= 0 ? 0 : errno;
#endif
}

/* on unix, this starts/stops clearcase views */
/* on windows (see the win specific platform.c) it also */
/* maps win shares */

int 	startview_timeout;
void startview_timeout_handler(int signum) {
	startview_timeout = 1;
}

/* look for external library functions in a dyn library */
#include <dlfcn.h>

#ifdef HAVE_DLFCN_H
typedef int (*bf_dynamic_t)(char *, int, int *);
int bf_dynamic_function(char *cmd, int timeout, int *result, int *timeout_occurred) {
	char    *buf, *p, *soname, *fname;
	void    *dlhandle;
	pid_t   pid, pstatus;
	int     fds[2];
	bf_dynamic_t func;

	if ((soname = bfconf_get("extensions", NULL)) == NULL)
		return 0;

	if ((dlhandle = dlopen(soname, RTLD_LAZY)) == NULL) {
		free(soname);
		return 0;
	}

	buf = strdup(cmd); /* use a working buffer */
	if ((p = strpbrk(buf, " \t\r\n")) != NULL) { /* split the first token off */
		*p++ = 0;
		for ( ; *p && isspace(((int)*p)) ; p++); /* jump over whitespace */
	}

	/* see if this function exists */
	dlerror();
	func = (bf_dynamic_t)dlsym(dlhandle, buf);
	if (dlerror() != NULL) {
		dlclose(dlhandle);
		free(buf);
		free(soname);
		return 0;
	}
	buf = realloc(buf, BFBUFSIZ);

	/* spin a child process to call the function. this lets us reset stdin/stdout/stderr */
	/* so the library can use printf to log data */
	if (pipe(fds) != 0) {
		format_message(buf, bfCat(E_EXEC_DL_NOPIPE), errno, buf, soname);
		DataBlock(buf, NULL);
		dlclose(dlhandle);
		free(soname);
		free(buf);
		return 1;
	}

	fname = strdup(buf);
	format_message(buf, bfCat(E_EXEC_DL_OK), fname, soname);
	free(fname);
	DataBlock(buf, NULL);
	pstatus = 0;
	switch (pid = fork()) {
		case -1:		/* failure */
			format_message(buf, bfCat(E_EXEC_DL_NOFORK), errno, buf, soname);
			DataBlock(buf, NULL);
			dlclose(dlhandle);
			free(soname);
			free(buf);
			return 1;

		case 0:			/* child */
			setpgid(getpid(), getpid());
			close(fds[0]);	/* close the read side of the communication channel */
			close(0);		/* close stdin */
			dup2(fds[1], 1);	/* set stdout */
			dup2(fds[1], 2);	/* set stderr */
			setbuf(stdout, NULL);
			setbuf(stderr, NULL);
			exit((*func)(p, timeout, timeout_occurred));

		default:		/* parent */
			child_pgrp = -pid;
			close(fds[1]);	/* close the write side of the communication channel */
			memset(buf, 0, BFBUFSIZ);
			while (relay_command_output(fds[0], 0)) {}
			relay_command_output(fds[0], 1);
			close(fds[0]);
			kill(pid, SIGTERM);
			waitpid(pid, &pstatus, 0);
			*result = WEXITSTATUS(pstatus);
			break;
	}
	dlclose(dlhandle);
	free(buf);
	free(soname);
	return 1;
}
#else
int bf_dynamic_function(char *cmd, int timeout, int *result, int *timeout_occurred) {
	return 0;
}
#endif /* HAVE_DLFCN_H */

int bf_LoadValue(void) {
	return 1;
}



#include "platform/sysconf_cpu.c"
#include "platform/sysconf_disk.c"
#include "platform/sysconf_mem.c"
#include "platform/sysconf_net.c"
#include "platform/sysconf_os.c"


/* This is a replacement for setenv if the OS doesn't provide it
 * (HP/UX and Solaris).  It uses the native encoding; if you want
 * UTF-8, use platform_setenv, instead.
 */
#ifndef HAVE_SETENV
#ifdef HAVE_PUTENV
#ifdef PREVENT_PUTENV_MEMORY_LEAK
static hash_ctx   *env_ctx = NULL;
static hash_table *env_ht  = NULL;
static void init_env_ht(void) {
	env_ctx = hash_ctx_new();
	env_ctx->free_value_fn = hash_ignore_free_fn;
	env_ht = hash_new(env_ctx, 89);
}
#endif /* PREVENT_PUTENV_MEMORY_LEAK */
#endif /* HAVE_PUTENV */
int bf_setenv(const char *name, const char *value, int overwrite)
{
	int ret;
	size_t nlen, vlen;
#ifdef HAVE_PUTENV
	char *buf, *p;

	if (NULL == name || !*name || NULL == value || NULL != (p = strchr(name, '='))) {
		errno = EINVAL;
		return -1;
	}
	if (!overwrite && NULL != (p = getenv(name))) {
		ret = 0;
	} else {
		nlen = strlen(name);
		vlen = strlen(value) + 1;
		buf = malloc(nlen + vlen + 2);
		p = &buf[nlen];

		memcpy(buf, name, nlen);
		*p++ = '=';
		memcpy(p, value, vlen);
		ret = putenv(buf);

#ifdef PREVENT_PUTENV_MEMORY_LEAK
		if (env_ht == NULL)
			init_env_ht();
		hash_put(env_ht, buf, nlen, NULL);
#endif /* PREVENT_PUTENV_MEMORY_LEAK */
	}
#else
	/* Unable to support environments at all on this platform?! */
	errno = EINVAL;
	ret = -1;
#endif /* HAVE_PUTENV */
	return ret;
}
#endif /* HAVE_SETENV */



/*
 * Implementations for platform.h.
 */

int platform_access(const char *pathname, int mode) {
	return access(convert_to_native(pathname), mode);
}


int platform_chdir(const char *path) {
	char *p = CleanUnixPath(strdup(path));
	int result = chdir(p);
	free(p);
	return result ? 1 : 0;
}


int platform_chmod(const char *path, int mode) {
	return chmod(convert_to_native(path), mode);
}


int platform_chown(const char *path, int uid, int gid) {
	return chown(convert_to_native(path), uid, gid);
}


/*
 * Note: On z/OS, we disguise UTF-8 data as EBCDIC by transcoding it.
 * Provided that __atoe_l and __etoa_l are perfect inverses (they should be),
 * this is a non-destructive transformation that makes it possible to
 * work with UTF-8 data directly on an EBCDIC system.  We have to give up
 * on it when we go to create files and the like (we have to convert to
 * the native encoding at some point, naturally), but it's as much as we
 * can do, for now.
 */
size_t platform_convert(const char *in, size_t inlen, char *out, size_t outlen, platform_convert_t mode)
{
	size_t outsize, len;
	iconv_t cd = ICONV_INVALID;
	char *o = out;
#ifdef ZOS
	char *zin = NULL;
#endif

	if (inlen == -1) inlen = strlen(in) + 1;
	if (inlen == 0) return 0;
	if (NULL == iconv_native_encoding)
		platform_setlocale("");

	switch (mode) {
		case CONVERT_NATIVE_TO_UTF8:
		case CONVERT_CONSOLE_TO_UTF8:
			cd = cd_native_to_utf8;
			break;
		case CONVERT_UTF8_TO_NATIVE:
		case CONVERT_UTF8_TO_CONSOLE:
			cd = cd_utf8_to_native;
			break;
	}

	/* If conversion is unnecessary, or if the iconv system fails to
	 * provide a suitable converter, then use the data as-is.  This is
	 * fine even on z/OS, as we would be converting between EBCDIC and
	 * UTF-8 escaped as EBCDIC, and falling back to ASCII when UTF-8
	 * can't be used because iconv doesn't work would leave us converting
	 * EBCDIC->ASCII->EBCDIC, which would be silly.
	 */
	if (ICONV_INVALID == cd) {
		len = (outlen < inlen) ? outlen : inlen;
		memcpy(out, in, len);
		return len;
	}

#ifdef ZOS
	/* If we are converting from UTF-8 to the native encoding, we
	 * first have to undo the transformation that escaped UTF-8 data
	 * as EBCDIC.
	 */ 
	if (mode == CONVERT_UTF8_TO_NATIVE ||
			mode == CONVERT_UTF8_TO_CONSOLE)
	{
		zin = malloc(inlen);
		memcpy(zin, in, inlen);
		__etoa_l(zin, inlen);
		in = zin;
	}
#endif

	/* Note: the (char **) cast for &in is only necessary because
	 * GNU iconv declares it incorrectly.  There should never be a
	 * need for iconv to alter the input sequence.  The warning on
	 * other platforms is of no consequence.
	 */
	outsize = outlen;
	while (-1 == iconv(cd, (char **)&in, &inlen, &o, &outlen)) {
		if (errno == EILSEQ) {
			++in;
			--inlen;
			continue;
		}

		/* Take no chances.  If anything could possibly have left the
		 * decoder in a bad state, then start over.  It isn't worth it
		 * to try to reset the decoder, especially since HP/UX's iconv
		 * tends to throw a SIGSEGV when you try ...
		 */
		bf_iconv_close();
		break;
	}
	len = outsize - outlen;

#ifdef ZOS
	/* If we are converting from the native encoding to UTF-8, we
	 * need to perform the transformation that escapes UTF-8 data
	 * as EBCDIC.  Otherwise, we need to release the buffer we
	 * allocated earlier for holding the transcoded input for
	 * going the other direction.
	 */
	if (mode == CONVERT_UTF8_TO_NATIVE ||
			mode == CONVERT_UTF8_TO_CONSOLE)
	{
		free(zin);
	} else {
		__atoe_l(out, len);
	}
#endif

	return len;
}


char *platform_default_config_file(void) {
	return strdup("/etc/bfagent.conf");
}


const char *platform_default_lang(void) {
	const char *locale = platform_setlocale(NULL);
	return (NULL != locale) ? locale : "en_US";
}


char *platform_fgets(char *s, int size, FILE *stream) {
	if (platform_utf8_is_native())
		return fgets(s, size, stream);

	if (NULL == fgets(scbuf, BFBUFSIZ, stream))
		return NULL;

	convert_native_to_utf8(scbuf, -1, s, size);
	s[size-1] = '\0';
	return s;
}


FILE *platform_fopen(const char *path, const char *mode) {
	return fopen(convert_to_native(path), mode);
}


char *platform_getcwd(char *buf, size_t len) {
	if (platform_utf8_is_native())
		return getcwd(buf, len);

	if (NULL == getcwd(scbuf, BFBUFSIZ))
		return NULL;

	convert_native_to_utf8(scbuf, -1, buf, len);
	buf[len - 1] = '\0';
	return buf;
}


char *platform_getenv(const char *name)
{
#ifdef HAVE_GETENV
	char *p, *ret;
	if (NULL == name || !*name || NULL != (p = strchr(name, '=')))
		return NULL;
	ret = getenv(convert_to_native(name));
	return ret;
#else
	/* Unable to support environments at all on this platform?! */
	return NULL;
#endif
}


const char *platform_hostname(void)
{
	static char	*hostname = NULL;
	if (NULL == hostname) {
		char buf[BFBUFSIZ];
		if (gethostname(buf, BFBUFSIZ-1) != 0)
			return "localhost";
		buf[BFBUFSIZ-1] = '\0';
		hostname = strdup(buf);
	}
	return hostname;
}


const char *platform_id(void) {
	static char osname[128];
	struct utsname ut;

	uname(&ut);
	LowerString(ut.sysname);
	sprintf(osname, "%s %s %s", ut.sysname, ut.release, ut.version);
	return osname;
}


int platform_mkdir(const char *path, int mode) {
	return mkdir(convert_to_native(path), mode);
}


int platform_open2(const char *path, int flags) {
	return open(convert_to_native(path), flags);
}


int platform_open3(const char *path, int flags, int mode) {
	return open(convert_to_native(path), flags, mode);
}


int platform_openpty(void) {
	bf_command_state_t *state = &command_state;

	if (bf_openpty() == 0) {
		format_message(state->buf, bfCat(E_EXEC_PTY), state->ptyname);
		DataBlock(state->buf, NULL);
		return 0;
	}

	format_message(state->buf,
			bfCat(E_EXEC_NOPTY),
			bfCat((state->mfd < 0) ? M_PTY_MASTER : M_PTY_SLAVE),
			errno);
	DataBlock(state->buf, NULL);
	ResultBlock(E_EXEC_NOPTY, -1, state->buf);
	return -1;
}


int platform_protocol_has_socket(void) {
	struct sockaddr_storage name;
	socklen_t len = (socklen_t)sizeof(name);
	int rc = getpeername(ofd, (struct sockaddr *)&name, &len);
	return 0 == rc || ENOTSOCK != errno;
}


int platform_protocol_read(void *buf, size_t len) {
	int left = (int)len;
	if (left <= 0) { return left; }
	return (int)read(ifd, buf, left);
}


int platform_protocol_write(const void *buf, size_t len) {
	const char *p = buf;
	int left = (int)len;
	int bytes;

	if (left <= 0) { return -1; }
	do {
		if ((bytes = (int)write(ofd, p, left)) <= 0)
			return -errno;
		p += bytes;
		left -= bytes;
	} while (left > 0);
	return len;
}


int platform_rmdir(const char *path) {
	return rmdir(convert_to_native(path));
}


int platform_set_user(const char *username, const char *password)
{
	int 	rc;
	struct	passwd *p;


#ifdef ISERIES_AGENT
	char qsycusrs_return;
	char qsygetph_return[12];
		
	const char qsycusrs_profile[10] = {0x5c, 0xc3, 0xe4, 0xd9, 0xd9, 0xc5, 0xd5, 0xe3, 0x40, 0x40};
			 /* *CURRENT in EBCDIC */
	const char qsycusrs_authority[10] = {0x5c, 0xc1, 0xd3, 0xd3, 0xd6, 0xc2, 0xd1, 0x40, 0x40, 0x40};
			/* *ALLOBJ */
	
	char qsygetph_username[32];
	char qsygetph_password[32];
	char qsygetph_temp[32];

	const int qsycusrs_numauth = 1;
	const int qsycusrs_calllevel = 1;
	const int qsygetph_ccsid = 37;
	
	int iseries_rc;
	int qsygetph_passlen;
	int i;
	
	struct {
		int bytes_provided;
		int bytes_available;
		char msgid[7];
	} qsycusrs_errcode, qsygetph_errcode, qwtsetp_errcode;

	void *qsycusrs_argv[7];
	void *qsygetph_argv[6];
	void *qwtsetp_argv[2];

	ILEpointer qsycusrs_pointer, qsygetph_pointer, qwtsetp_pointer;


	/*  First, we call QSYCUSRS to check if the current process has the *ALLOBJ special 
	 *  authority which is required to change identity.
	 */
	memset(&qsycusrs_errcode, 0, sizeof(qsycusrs_errcode));
	memset(&qsycusrs_pointer, 0, sizeof(qsycusrs_pointer));
	memset(&qsycusrs_argv, 0, sizeof(qsycusrs_argv));

	qsycusrs_errcode.bytes_provided = sizeof(qsycusrs_errcode);

	qsycusrs_argv[0] = &qsycusrs_return;
	qsycusrs_argv[1] = &qsycusrs_profile;
	qsycusrs_argv[2] = &qsycusrs_authority;
	qsycusrs_argv[3] = (void *) &qsycusrs_numauth;
	qsycusrs_argv[4] = (void *) &qsycusrs_calllevel;
	qsycusrs_argv[5] = &qsycusrs_errcode;
	qsycusrs_argv[6] = NULL;

	iseries_rc = _RSLOBJ2(&qsycusrs_pointer,
					  RSLOBJ_TS_PGM,
					  "QSYCUSRS",
					  "QSYS");

	iseries_rc = _PGMCALL(&qsycusrs_pointer, 
					  (void*)&qsycusrs_argv, 
					  0);

	/* save the existing id bits */
	original_uid = getuid();
	original_gid = getgid();

	/* if this agent is not running as QSECOFR or doesn't have *ALLOBJ
	 * special authority, then check_magic_login applies, and you must 
	 * specify login information if that is set,
	 * or must not if it is not set.
	 */
	if (original_uid && qsycusrs_return != '\xE8' /* 'Y' (IBM-037) */) {
		return check_magic_login(username, password);
	}

	/* If this agent is running as root, then you *must* supply a valid
	 * username and password.
	 */
	if (!username || !password || !*username || !*password)  {
		return SU_ILLEGAL;
	}

	memset(&qsygetph_errcode, 0, sizeof(qsygetph_errcode));
	
	/*  The username needs to be an uppercase space-padded string of 10 EBCDIC characters. */
	strncpy(qsygetph_temp, username, 10);
	qsygetph_temp[10] = 0;
	strcat(qsygetph_temp, "         ");
	qsygetph_temp[10] = 0;
	
	for (i = 0; i < 10; i++)  {
		qsygetph_temp[i] = toupper(qsygetph_temp[i]);
	}

	/*  Converting username and password to EBCDIC. */
	iconv_t cd = iconv_open("IBM-037", "UTF-8");
	if ((int) cd == -1)  {
		printf("iconv_open failed - errno = %d - ", errno);
		switch (errno)  {
			case EFAULT:  printf("EFAULT\n");  break;
			case EINVAL:  printf("EINVAL\n");  break;
			case ENOMEM:  printf("ENOMEM\n");  break;
			default:  printf("EUNKNOWN\n");  break;
		}
		return SU_ILLEGAL;
	}

	size_t iconv_templen = strlen(qsygetph_temp) + 1;
	size_t iconv_userlen = sizeof(qsygetph_username);
	const char *in = qsygetph_temp;
	char *out = qsygetph_username;
	iconv(cd, (char **)&in, &iconv_templen, &out, &iconv_userlen);

	size_t iconv_passlen = sizeof(qsygetph_password);
	iconv_templen = sizeof(qsygetph_password);
	in = password;
	out = qsygetph_password;
	iconv(cd, (char **)&in, &iconv_passlen, &out, &iconv_templen);
	iconv_close(cd);


	/*  Calling QSYGETPH to return a process handle with the associated user 
	 *  permissions.  Only doable if we have the proper login and password.
	 */
	qsygetph_passlen = strlen(qsygetph_password);
	
	qsygetph_errcode.bytes_provided = sizeof(qsygetph_errcode);

	qsygetph_argv[0] = &qsygetph_username;
	qsygetph_argv[1] = &qsygetph_password;
	qsygetph_argv[2] = &qsygetph_return;
	qsygetph_argv[3] = &qsygetph_errcode;
	qsygetph_argv[4] = &qsygetph_passlen;
	qsygetph_argv[5] = (void *) &qsygetph_ccsid;

	iseries_rc = _RSLOBJ2(&qsygetph_pointer,
					  RSLOBJ_TS_PGM,
					  "QSYGETPH",
					  "QSYS");

	iseries_rc = _PGMCALL(&qsygetph_pointer, 
					  (void*)&qsygetph_argv, 
					  0);

	if (qsygetph_errcode.bytes_available)  {
		return SU_ILLEGAL;
	}
	

	/*  Calling QWTSETP to set the current process to use the previously generated
	 *  process handle.
	 */
	memset(&qwtsetp_errcode, 0, sizeof(qwtsetp_errcode));

	qwtsetp_errcode.bytes_provided = sizeof(qwtsetp_errcode);

	qwtsetp_argv[0] = &qsygetph_return;
	qwtsetp_argv[1] = &qwtsetp_errcode;

	iseries_rc = _RSLOBJ2(&qwtsetp_pointer,
					  RSLOBJ_TS_PGM,
					  "QWTSETP",
					  "QSYS");

	iseries_rc = _PGMCALL(&qwtsetp_pointer, 
					  (void*)&qwtsetp_argv, 
					  0);

	if (qwtsetp_errcode.bytes_available)  {
		return SU_SETUSERFAILED;
	}
	return SU_OK;
# else /* !ISERIES_AGENT */

	/* save the existing id bits */
	original_uid = getuid();
	original_gid = getgid();
#endif /* ISERIES_AGENT/! */

	/* if this agent is not running as root, then check_magic_login
	 * applies, and you must specify login information if that is set,
	 * or must not if it is not set.
	 */
	if (original_uid)
		return check_magic_login(username, password);

	/* If this agent is running as root, then you *must* supply a valid
	 * username and password.
	 */
	if (!username || !password || !*username || !*password)
		return SU_ILLEGAL;

	/* Start a PAM session using "bfagent" as the program name.  If "bfagent"
	 * doesn't match, then try to fall back on "login".
	 */
	if ((rc = valid_credentials("bfagent", username, password))) {
		/* but not if bfagent is in pam and the auth failed for some other reason */
		if (rc != SU_PAMSESSIONERROR)
			return rc;
		if ((rc = valid_credentials("login", username, password)))
			return rc;
	}

	/* do setuid code here. */
	if ((p = getpwnam(username)) == NULL)
		return SU_USERNOTFOUND;

	setgid(p->pw_gid);
#if HAVE_INITGROUPS
	initgroups(username, getgid());
#endif

	if (setuid(p->pw_uid) != 0) {
		endpwent();
		return SU_SETUSERFAILED;
	}

	endpwent();
	return SU_OK;
}


int platform_setenv(const char *name, const char *value, int overwrite)
{
	char *n, *v;
	size_t nlen, vlen;
	int ret;

	if (NULL == name || !*name || NULL == value) {
		errno = EINVAL;
		return -1;
	}
	nlen = strlen(name) + 1;
	n = malloc(nlen << 1);
	convert_utf8_to_native(name, nlen, n, nlen << 1);
	vlen = strlen(value) + 1;
	v = malloc(vlen << 1);
	convert_utf8_to_native(value, vlen, v, vlen << 1);
	ret = setenv(n, v, overwrite);

	free(v);
	free(n);
	return ret;
}

const char *platform_setlocale(const char *locale) {
#ifdef HAVE_LOCALE_H
	if (NULL != locale) {
		if (NULL == setlocale(LC_ALL, locale))
			return NULL;
		bf_iconv_close();
		bf_iconv_open();
	}
	return setlocale(LC_CTYPE, NULL);
#else
	errno = -1;
	return NULL;
#endif
}

void platform_signals_catch(const char *who) {
	if (NULL != signal_catcher) {
		free(signal_catcher);
	}
	signal_catcher = strdup(who);
	signal(SIGTERM, fatal_signal);
	signal(SIGINT, fatal_signal);
	signal(SIGQUIT, fatal_signal);
	signal(SIGHUP, ignore_signal);
	signal(SIGPIPE, fatal_signal);
	signal(SIGCHLD, reap_zombies);
}

void platform_signals_release(void) {
	if (NULL != signal_catcher) {
		free(signal_catcher);
	}
	signal_catcher = NULL;
	signal(SIGTERM, SIG_DFL);
	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	signal(SIGPIPE, SIG_DFL);
	signal(SIGCHLD, SIG_DFL);
}

int platform_stat(const char *path, struct stat *buf) {
	return stat(convert_to_native(path), buf);
}


platform_style_t platform_style(void) {
	return STYLE_UNIX;
}


/* We normally put tmp files in /tmp, but you can set "tmpdir" in the
 * conf file to make them go elsewhere.
 */
int platform_tempfd(char **namebuf) {
	char *p = bfconf_get("tmpdir", "/tmp");
	char *buf = NULL;
	int fd;

#ifdef HAVE_MKSTEMP
	buf = calloc(1, strlen(p) + 16);
	sprintf(buf, "%s/bfXXXXXX", p);
	CleanPath(buf);
	fd = mkstemp(buf);
#else /* !HAVE_MKSTEMP */
	while (1) {
		/* if tempnam pukes we'll create a time/pid based name and give up */
		if (NULL == (buf = tempnam(p, NULL))) {
			buf = calloc(1, MAX_PATH);
			sprintf(buf, "%s/%ld.%d.tmp", p, (long)time(NULL), (int)getpid());
		}

		/* make sure the name is unique */
		if ((fd = open(buf, O_CREAT|O_EXCL|O_RDWR, S_IRUSR|S_IWUSR)) >= 0)
			break;
	}
#endif /* HAVE_MKSTEMP */

	free(p);
	if (NULL == namebuf) {
		free(buf);
	} else {
		*namebuf = buf;
	}
	return fd;
}


/* this is basically like platform_tempfd() but returns the name */
/* instead. */
char *platform_tempnam(void) {
	char *result;
	platform_close(platform_tempfd(&result));
	return result;
}


int platform_unlink(const char *pathname) {
	return unlink(convert_to_native(pathname));
}


void platform_unmap_drives(void) {
	if (mapStr) {
		free(mapStr);
		mapStr = NULL;
	}
}


int platform_unsetenv(const char *name)
{
	char *n, *p;
	size_t nlen;
#if !defined(HAVE_SETENV)
	char **e;
#endif

	if (NULL == name || !*name || NULL != (p = strchr(name, '='))) {
		errno = EINVAL;
		return -1;
	}

	nlen = strlen(name) + 1;
	n = malloc(nlen << 1);
	nlen = convert_utf8_to_native(name, nlen, n, nlen << 1);

#ifdef HAVE_UNSETENV
	unsetenv(n);
#else
	n[nlen++] = '=';
	n[nlen] = '\0';
	for (e = environ; *e; e++) {
		if (strncmp(*e, n, nlen) == 0) {
			do {
				*e = *(e+1);
				++e;
			} while (*e);
#ifdef PREVENT_PUTENV_MEMORY_LEAK
			if (env_ht != NULL)
				hash_del(env_ht, n, nlen-1);
#endif /* PREVENT_PUTENV_MEMORY_LEAK */
			break;
		}
	}
#endif

	free(n);
	return 0;
}

int platform_utime(const char *pathname, struct utimbuf *times) {
	return utime(convert_to_native(pathname), times);
}

char *platform_update_path(void) {
	return bfconf_get("update_path", "/usr/local/bin/bfagent");
}


#ifndef HAVE_FREE
#undef free
void rpl_free(void *ptr) {
	if (NULL != ptr)
		free(ptr);
}
#endif /* !def HAVE_FREE */


#ifndef HAVE_MALLOC
#undef malloc
void *rpl_malloc(size_t len) {
	return malloc(len > 0 ? len : 1);
}
#endif /* !def HAVE_MALLOC */


#ifndef HAVE_REALLOC
#undef realloc
void *rpl_realloc(void *ptr, size_t len) {
	if (ptr == NULL)
		return malloc(len > 0 ? len : 1);
	if (len != 0)
		return realloc(ptr, len);
	free(ptr);
	return NULL;
}
#endif /* !def HAVE_REALLOC */

