/*
 * Copyright (C)2003-2008, IBM, Inc. All rights reserved.
 * IBM, Inc. owns the copyright and other intellectual
 * property rights in this software. Duplication or use of the
 * Software is not permitted except as expressly provided in a
 * written agreement between your company and IBM, Inc.
 */

/* AIX -- make sure we have an implementation for vmgetinfo */
#ifdef HAVE_SYS_VMINFO_H /* AIX */
#ifndef HAVE_VMGETINFO
#ifdef HAVE_DLFCN_H
static int (*vmgetinfo_ksym)(void *, int, int) = NULL;
static int bf_vmgetinfo(void *out, int command, int arg) {
	if (vmgetinfo_ksym == NULL) {
		void *ksyms = dlopen("/unix", RTLD_NOW | RTLD_GLOBAL);
		if (ksyms == NULL)
			return -1;

		*(void **)(&vmgetinfo_ksym) = dlsym(ksyms, "vmgetinfo");
		dlclose(ksyms);
		if (vmgetinfo_ksym == NULL)
			return -1;
	}
	return (*vmgetinfo_ksym)(out, command, arg);
}
#define vmgetinfo(out, command, arg) (bf_vmgetinfo(out, command, arg))
#else /* !HAVE_DLFCN_H */
#define vmgetinfo(out, command, arg) (-1)
#endif /* HAVE_DLFCN_H */
#endif /* !HAVE_VMGETINFO */
#endif /* HAVE_SYS_VMINFO_H */


static void get_mem_info(platform_sysconf_mem_t *mem)
{
#ifdef HAVE_MACH_MACH_H /* MacOS */
	mach_port_t port;
	unsigned int count = HOST_VM_INFO_COUNT;
	vm_statistics_data_t vm;

	mem->mem_total = mem->mem_used = mem->mem_free = mem->mem_load = -1.0;

	port = mach_host_self();
	if (host_statistics(port, HOST_VM_INFO, (host_info_t)&vm, &count) == KERN_SUCCESS) {
		mem->mem_used = (double)(vm.active_count + vm.inactive_count + vm.wire_count) * mem->mem_page_size;
		mem->mem_free = (double)vm.free_count * mem->mem_page_size;
		mem->mem_total = mem->mem_used + mem->mem_free;
		mem->mem_load = mem->mem_used / mem->mem_total;
		return;
	}
	/* Fall back on sysconf */
#elif defined(HAVE_SYS_VMINFO_H) /* AIX */
	struct vminfo vm;

	if (vmgetinfo(&vm, VMINFO, sizeof(vm)) == 0) {
		mem->mem_used = (double)(vm.numwseguse + vm.numpseguse + vm.numclseguse) * mem->mem_page_size;
		mem->mem_free = (double)vm.numfrb * mem->mem_page_size;
		mem->mem_total = (double)vm.memsizepgs * mem->mem_page_size;
		mem->mem_load = mem->mem_used / mem->mem_total;
		return;
	}
	/* Fall back on sysconf */
#endif /* MacOS or AIX */

	/* Try to get the memory usage info from sysconf */
#ifdef _SC_AVPHYS_PAGES
	mem->mem_free = (double)sysconf(_SC_AVPHYS_PAGES) * mem->mem_page_size;
#endif
#ifdef _SC_PHYS_PAGES
	mem->mem_total = (double)sysconf(_SC_PHYS_PAGES) * mem->mem_page_size;
	if (mem->mem_free >= 0.0f && mem->mem_total >= mem->mem_free) {
		mem->mem_used = mem->mem_total - mem->mem_free;
		mem->mem_load = mem->mem_used / mem->mem_total;
	}
#endif
}

platform_sysconf_mem_t *platform_sysconf_mem_new(void) {
	platform_sysconf_mem_t *mem = calloc(1, sizeof(platform_sysconf_mem_t));
	long mem_page_size = 4096;
	char *p;

	if (NULL == mem)
		return NULL;

	if (NULL != ( p = bfconf_get("memory_page_size", NULL) )) {
		int i = atoi(p);
		if (i > 0)
			mem_page_size = i;
		free(p);
	}

#ifdef _SC_PAGE_SIZE
	{
		long value;
		if ((value = sysconf(_SC_PAGE_SIZE)) > 0)
			mem_page_size = value;
	}
#endif
	if (mem_page_size <= 0)
		mem_page_size = 4096;

	mem->mem_page_size = mem_page_size;
	get_mem_info(mem);
	return mem;
}

