/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int RemoveFunction(char *, struct stat *, int);

int FuncRmdir(void)
{
	char	*root, *dir, *odir, buf[BFBUFSIZ], msg[BFBUFSIZ], *userdata;
	char	*ccroot, *timeout;
	int 	ret = 0, mapfail;
	size_t	userdata_size;

	/*  locate the root element */
	if (NULL == (root = request_get(TAG_ROOT))) {
		ResultBlock(E_RMDIR_NOROOT, -1, bfCat(E_RMDIR_NOROOT));
		return 0;
	}

	/*  locate the dir element */
	if (NULL == (odir = request_get(TAG_DIR))) {
		ResultBlock(E_RMDIR_NODIR, -1, bfCat(E_RMDIR_NODIR));
		return 0;
	}

	/* don't allow ..'s in dir */
	if (strstr(odir, "..")) {
		ResultBlock(E_RMDIR_DOTDOT, -1, bfCat(E_RMDIR_DOTDOT));
		return 0;
	}

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the client timeout element */
	if (NULL == (timeout = request_get(TAG_TIMEOUT)))
		timeout = "0";

	/*  locate the silent element */
	verbose = !request_has(TAG_SILENT);

	StoreUserData(userdata, userdata_size);

	DumpEnv();

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		ResultBlock(E_EXEC_MAPFAIL, -1, bfCat(E_EXEC_MAPFAIL));
		FreeUserData();
		return 0;
	}

	ClobberPasswords();

	CleanPath(root);
	CleanPath(odir);

	root = env_expand_var(root, 0, 0);
	dir = env_expand_var(odir, 0, 0);
	CleanPath(dir);
	if (strcmp(dir, odir)) {
		format_message(msg, bfCat(E_RMDIR_XLATE), root, odir, root, dir);
		CleanPath(msg);
		if (verbose)
			DataBlock(msg, NULL);
	}

	ccroot = NULL;
	if (clearcase_view) {
		ccroot = calloc(1, BFBUFSIZ);
		format_message(ccroot, "%s/%s/%s", this_view(), clearcase_view, root);
		root = ccroot;
	}

	/* build the full base path that we'll be removing */
	format_message(buf, "%s/%s", root, dir);
	CleanPath(buf);

	/*  test to see if it doesn't exist already */
	if (0 == platform_access(buf, 0)) {
		/* buf now has the path to delete files from */
		/* the actual work is off in platform.c 'cause some systems don't have */
		/* nftw (?) and rm -rf doesn't work in windows */
		BFRemoveDirectory(buf);

		/*  test to see if it's still there */
		if (0 == platform_access(buf, 0))
			ret = E_RMDIR_FAIL;
	}


	platform_unmap_drives();
	format_message(msg, bfCat(ret == 0 ? E_RMDIR_OK : E_RMDIR_FAIL), buf);
	FreeUserData();
	ResultBlock(ret, -1, msg);
	free(dir);
	if (ccroot)
		free(ccroot);
	return 0;
}

