/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int FuncUnlink(void)
{
    char    *root, *path, *opath, buf[BFBUFSIZ], msg[BFBUFSIZ], *userdata;
    char    *ccroot, *timeout;
    int     ret = E_UNLINK_OK, mapfail;
    size_t  userdata_size;

    /*  locate the root element */
    if (NULL == (root = request_get(TAG_ROOT))) {
        ResultBlock(E_UNLINK_NOROOT, -1, bfCat(E_UNLINK_NOROOT));
        return 0;
    }

    /*  locate the path element */
    if (NULL == (opath = request_get(TAG_PATH))) {
        ResultBlock(E_UNLINK_NOPATH, -1, bfCat(E_UNLINK_NOPATH));
        return 0;
    }
    /* don't allow ..'s in path */
    if (strstr(opath, "..")) {
        ResultBlock(E_UNLINK_DOTDOT, -1, bfCat(E_UNLINK_DOTDOT));
        return 0;
    }

    /*  locate the userdata element */
    userdata = request_get_binary(TAG_USERDATA, &userdata_size);

    /*  locate the client timeout element */
    if (NULL == (timeout = request_get(TAG_TIMEOUT)))
        timeout = "0";

    /*  locate the silent element */
    verbose = !request_has(TAG_SILENT);

    StoreUserData(userdata, userdata_size);

    DumpEnv();

    mapfail = BFCONF_EXISTS(map_drive_is_failure);
    if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
        ResultBlock(E_EXEC_MAPFAIL, -1, bfCat(E_EXEC_MAPFAIL));
        FreeUserData();
        return 0;
    }

	ClobberPasswords();

    CleanPath(root);
    CleanPath(opath);

    root = env_expand_var(root, 0, 0);
    path = env_expand_var(opath, 0, 0);
    CleanPath(path);
    if (strcmp(path, opath)) {
        format_message(msg, bfCat(E_UNLINK_XLATE), root, opath, root, path);
        CleanPath(msg);
        if (verbose)
            DataBlock(msg, NULL);
    }
    ccroot = NULL;
    if (clearcase_view) {
        ccroot = calloc(1, BFBUFSIZ);
        format_message(ccroot, "%s/%s/%s", this_view(), clearcase_view, root);
        root = ccroot;
    }

    /* build the full base path that we'll be removing */
    format_message(buf, "%s/%s", root, path);
    CleanPath(buf);

    /*  test to see if it doesn't exist already */
    if (0 == platform_access(buf, 0)) {
		/* check to see if it's a dir */
		if (isDir(buf)) {
			ret = E_UNLINK_DIR;
		} else {
			/* all good. blow it away */
			platform_unlink(buf);

			/*  test to see if it's still there */
			ret = (0 == platform_access(buf, 0))
				? E_UNLINK_FAIL
				: E_UNLINK_OK;
		}
	}

    platform_unmap_drives();
	format_message(msg, bfCat(ret), buf);
    FreeUserData();
    ResultBlock(ret == E_UNLINK_FAIL ? E_UNLINK_FAIL : 0, -1, msg);
    free(path);
    if (ccroot)
        free(ccroot);
    return 0;
}

