/*
Copyright (C)2003-2007, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

int FuncWrite(void)
{
	char	*buf = malloc(BFBUFSIZ), *path, *data = NULL, *p, *userdata;
	char	*timeout;
	int 	bytes, total, len, fd = -1, mode, append, clobber;
	time_t	stamp;
	struct 	utimbuf tims;
	int     mapfail, rc = 0, arg = -1;
	int     stream_fd = -1;
	off_t   offset = (off_t)0;
	size_t	userdata_size;

	/*  locate the path element */
	if ((path = request_get(TAG_PATH)) == NULL) {
		ResultBlock(E_WRITE_NOPATH, -1, bfCat(E_WRITE_NOPATH));
		free(buf);
		return 0;
	}

	stamp = 0;
	if (NULL != (p = request_get(TAG_STAMP)))
		stamp = atol(p);

	/* See if a data stream was provided */
	if (-1 == (stream_fd = request_get_stream(TAG_STREAM))) {
		/* If the stream tag is not supplied, then this must be the old,
		 * inefficient style that uses a small data block and offset.
		 * The data is required.  The offset will default to 0.
		 */
		data = request_get(TAG_DATA);
		if (NULL == data) {
			format_message(buf, bfCat(E_WRITE_NODATA), path);
			ResultBlock(E_WRITE_NODATA, -2, buf);
			free(buf);
			return 0;
		}
		p = request_get(TAG_OFFSET);
		if (NULL != p) {
			while (isdigit((int)*p)) {
				offset = offset * 10 + (*p++ - '0');
			}
			if (offset < 0) { offset = 0; }
		}
	}

	/*  locate the append element. if it wasn't supplied, set a default. */
	append = request_has(TAG_APPEND) ? O_APPEND : 0;

	/*  locate the clobber element. if it wasn't supplied, set a default. */
	clobber = request_has(TAG_CLOBBER);

	/*  locate the mode element. if it wasn't supplied, set a default. */
	mode = -1;
	if (NULL != (p = request_get(TAG_MODE)))
		sscanf(p, "%o", (unsigned int *)&mode);

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the client timeout element */
	if (NULL == (timeout = request_get(TAG_TIMEOUT)))
		timeout = "0";

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ((platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail))
	{
		ResultBlock(E_EXEC_MAPFAIL, -1, bfCat(E_EXEC_MAPFAIL));
		FreeUserData();
		free(buf);
		return 0;
	}

	ClobberPasswords();

	if (clearcase_view) {
		char *view = this_view();
		p = calloc(1, strlen(view) + strlen(clearcase_view) +
				strlen(path) + 3);
		format_message(p, "%s/%s/%s", view, clearcase_view, path);
		path = p;
	} else {
		path = strdup(path);
	}
	/* Note: Do not use "goto done;" to bail out until after this
	 * point, as it would try to free(path);
	 */

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/* assemble and clean the final pathname. clean the
	 * pieces so we can use them individually later. */
	CleanPath(path);
	path = env_expand_var((p = path), 0, 0);
	free(p);

	/* clobber the target file?
	 * TODO: Error handling would be nice...
	 */
	if (clobber)
		platform_unlink(path);

	/* open the target file */
	total = 0;
	if (!clobber && 0 == platform_access(path, 0)) {
		if ((fd = platform_open2(path, O_WRONLY|O_BINARY|append)) >= 0) {
			if (offset > 0)
				lseek(fd, offset, SEEK_SET);
		}
	} else {
		fd = platform_open3(path, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY|append, 0600);
	}
	if (fd < 0) {
		format_message(buf, bfCat(E_WRITE_NOOPEN), path);
		rc = E_WRITE_NOOPEN;
		arg = -3;
		goto done;
	}

	if (stream_fd < 0) {
		/* In data block mode, we originally have several lines of hex
		 * in TAG_DATA.  We used to break it into the individual lines
		 * and decode and write out each line individually.  That was
		 * horribly inefficient.  This is a little bit better, but the
		 * conversion to streamed mode will be significantly better.
		 */
		char *p = buf;
		len = strlen(data) >> 1;
		if (len > BFBUFSIZ)
			p = buf = realloc(buf, len);

		len = DecodeBlock((unsigned char *)buf, (unsigned char *)data);
		while (len > 0) {
			if ((bytes = write(fd, p, len)) <= 0) {
				format_message(buf, bfCat(E_WRITE_WRITEFAIL), errno);
				rc = E_WRITE_WRITEFAIL;
				goto done;
			}
			total += bytes;
			p += bytes;
			len -= bytes;
		}
	}
	else {
		/* In stream mode, the data has already been decoded and stored in
		 * a tmp file; we just need to copy it.
		 */
		lseek(stream_fd, 0, SEEK_SET);
		while ((len = read(stream_fd, buf, BFBUFSIZ)) > 0) {
			char *p = buf;
			while (len > 0) {
				bytes = write(fd, p, len);
				if (bytes <= 0) {
					format_message(buf, bfCat(E_WRITE_WRITEFAIL), errno);
					lseek(stream_fd, 0, SEEK_END);
					rc = E_WRITE_WRITEFAIL;
					goto done;
				}
				total += bytes;
				p += bytes;
				len -= bytes;
			}
		}
	}
	close(fd);
	fd = -1;

	/* Set the file's permissions if a specific value was requested. */
	if (mode != -1) {
		if (0 == platform_chmod(path, mode))
			format_message(buf, bfCat(E_WRITE_SETMODE), path, mode);
		else
			format_message(buf, bfCat(E_WRITE_NOSETMODE), path, mode);
		DataBlock(buf, NULL);
	}

	/* Set the file's timestamp if a specific value was requested. */
	if (stamp > 0) {
		tims.actime = tims.modtime = stamp;
		if (0 == platform_utime(path, &tims))
			format_message(buf, bfCat(E_WRITE_SETTIME), path, ctime(&stamp));
		else
			format_message(buf, bfCat(E_WRITE_NOSETTIME), path, ctime(&stamp));
		DataBlock(buf, NULL);
	}

	format_message(buf, bfCat(E_WRITE_OK), total, path);

done:
	if (fd >= 0)
		close(fd);
	platform_unmap_drives();
	ResultBlock(rc, arg, buf);
	FreeUserData();
	free(path);
	free(buf);
	return 0;
}

